# -*- coding: utf-8 -*-
"""
Created 2015

@author: H. Rotzinger, MP Pfirrmann, M Wildermuth


"""
import logging
import h5py
import numpy as np
import qkit
from qkit.storage.hdf_constants import ds_types
from distutils.version import LooseVersion

file_kwargs = dict()
if LooseVersion(h5py.__version__) >= LooseVersion("3.5.0"): # new file locking
    file_kwargs = dict(locking=False)
elif LooseVersion(h5py.__version__) >= LooseVersion("3.0.0"): # intermediate
    logging.error("Qkit HDF file handling: In h5py between 3.0 and 3.5, there are problems with file locking handling. Please update to h5py==3.5.0")

class H5_file(object):
    """Base hdf5 class intended for qkit.
    
    hdf5 based Data class as compatible as possible to the standard data class. 
    It is in many respects more restricted to the hdf5_data.py class, 
    it e.g. does only create one group in the HDF5 file, etc.
    
    The class itself contans only the interface to the hdf5 file and 
    can be used also standalone.
    Here we again wrap the create_dataset() fctn and the append() fct does the
    trick of placing added data in the correct position in the dataset.
    """    
    
    def __init__(self,output_file, mode,**kw):
        """Inits the H5_file at the path 'output_file' with the access mode
        'mode'
        """
        self.create_file(output_file, mode)
        self.newfile = False
        
        if self.hf.attrs.get("qt-file",None) or self.hf.attrs.get("qkit",None):
            "File existed before and was created by qkit."
            self.setup_required_groups()
        else:
            "new file or none qkit file"
            self.newfile = True
            self.set_base_attributes()
        
            # set all standard attributes
            for k in kw:
                self.grp.attrs[k] = kw[k]
        
    def create_file(self,output_file, mode):
        self.hf = h5py.File(output_file, mode,**file_kwargs )

    def set_base_attributes(self):
        "stores some attributes and creates the default data group"
        # store version of the file format
        self.hf.attrs.create("qkit", "1.0".encode())  # qkit version
        # make the structure compatible with the nexus format
        # maybe some day the data can by analyzed by the software supporting nexus
        # first the entry group
        self.hf.attrs.create("NeXus_version","4.3.0".encode())
        
        self.entry = self.hf.require_group("entry")
        self.entry.attrs.create("NX_class","NXentry".encode())
        self.entry.attrs.create("data_latest",0)
        self.entry.attrs.create("analysis_latest",0)
        self.entry.attrs.create("updating",True)
        # create a nexus data group        
        self.dgrp = self.entry.require_group("data0")
        self.agrp = self.entry.require_group("analysis0")
        self.vgrp = self.entry.require_group("views")
        self.dgrp.attrs.create("NX_class","NXdata".encode())
        self.dgrp.attrs.create("NX_class","NXdata".encode())
    
    def setup_required_groups(self):
        # make the structure compatible with the nexus format
        # maybe some day the data can by analyzed by the software supporting nexus
        # first the entry group
        self.entry = self.hf.require_group("entry")
        # create a nexus data group        
        self.dgrp = self.entry.require_group("data0")
        self.agrp = self.entry.require_group("analysis0")
        self.vgrp = self.entry.require_group("views")
        
    def create_dataset(self,name, tracelength, ds_type = ds_types['vector'],
                       folder = "data", dim = 1, **kwargs):
        """Dataset for one, two, and three dimensional data
        
            Args:
                
                'tracelength'
            
                'dim'
                    is 1 for a single data point in a 1D scan (array of scalars)
                    is the length of the first trace in 2D scan (array of vectors)
                    For 2D scans the traces have to have the same tracelength
                    and are simply appended to the trace array
            
                'folder' is a optional group relative to the default group
            
                'kwargs' are appended as attributes to the dataset
        """

        self.ds_type = ds_type
        
        if dim == 1:
            shape    = (0,)
            maxshape = (None,)
            chunks=True
            
        elif dim == 2:
            shape    = (0,0)
            maxshape = (None,None)
            chunks=(5, tracelength)
            
        elif dim == 3:
            shape    = (0,0,0)
            maxshape = (None,None,None)
            chunks=(5, 5, tracelength)
            
        else:
            logging.error("Create datasets: '%s' is wrong number of dims." %(dim))
            raise ValueError

        if folder == "data":
            self.grp = self.dgrp
        elif folder == "analysis":
            self.grp = self.agrp
        elif folder == "views":
            self.grp = self.vgrp

        else:
            logging.error("please specify either: folder = 'data' , folder = 'analysis' or folder ='view' ")
            raise ValueError
            
        if name in self.grp.keys():
            logging.info("Item '%s' already exists in data set." % (name))
            #return False        
            
        # by default we create float datasets        
        dtype = kwargs.get('dtype','f')
        
        # we store text as unicode; this seems somewhat non-standard for hdf
        if ds_type == ds_types['txt']:
            try:
                dtype = h5py.special_dtype(vlen=unicode) # python 2
            except NameError:
                dtype = h5py.special_dtype(vlen=str) # python 3
        #create the dataset ...;  delete it first if it exists, unless it is data
        if name in self.grp.keys(): 
            if folder == "data":
                logging.error("Dataset '%s' in 'data' already exists. Cannot overwrite datasets in 'data'!" % (name))
                raise ValueError
            else:
                del self.grp[name]
                # comment: The above line does remove the reference to the dataset but does not free the space aquired
                # fixme if possible ...


        # 'scaleoffset' is an optional parameter for lossy compression of floating-point data,  retaining a specified number of bits post-decimal. 
        # It is used to compress dataset elements by reducing the precision of the data. Defaults to None, implying no compression.
        scaleoffset = kwargs.get('scaleoffset',None)

        if ds_type == ds_types['txt']:
            ds = self.grp.create_dataset(name, shape, maxshape=maxshape, chunks = chunks, dtype=dtype, scaleoffset = scaleoffset)
        else:
            ds = self.grp.create_dataset(name, shape, maxshape=maxshape, chunks = chunks, dtype=dtype, fillvalue = np.nan, scaleoffset = scaleoffset)
        
        ds.attrs.create("name",name.encode())
        ds.attrs.create("ds_type", ds_type)
        if ds_type == ds_types['matrix'] or ds_type == ds_types['box']:
            ## fill value only needed for >1D datasets
            ds.attrs.create("fill", [0,0,0])
        # add attibutes
        for a in kwargs:
            if not a == "scaleoffset":
                ds.attrs.create(a,(kwargs[a]).encode())
             
        self.flush()
        return ds
        
    def append(self,ds,data, next_matrix=False, reset=False, pointwise=False):
        """Method for appending hdf5 data. 
        
        A simple append method for data traces.
        Reshapes the array and updates the attributes.
        The optional 'next_matrix' atrribute arranges the incoming data in a 
        value_box correctly.
        
        Args:
            hdf_dataset 'ds'
            numpy array 'data'
            boolean 'next_matrix'
            pointwise (Boolean): if True, the data is appended pointwise, i.e. to the innermost dimension
        Returns:
            The function operates on the given variables.
        """
        # it gets a little ugly with all the different user-cases here ...
        if len(ds.shape) == 1:
            ## 1dim dataset (text, coordinate, vector)
            ## multiple inputs: text, scalar (not needed?), list/np.array with one or multiple entries
            if self.ds_type == ds_types['txt']:
                ## text
                dim1 = ds.shape[0]+1
                ds.resize((dim1,))
                ds[dim1-1] = data
                """
                ## This should not be needed anymore as there are non-user typecasts to np arrays before calling this fct.
                elif len(data.shape) == 0:              
                    ## scalar input; this is more or less useless, all non-text data gets cast
                    ## into np.array prior to calling this function.
                    dim1 = ds.shape[0]+1
                    fill[0] += 1
                    ds.resize((dim1,))
                    ds[fill[0]-1] = data
                """
            elif len(data.shape) == 1:
                ## np array (or list)
                if len(data) == 1:                  
                    ## single entry
                    dim1 = ds.shape[0]+1
                    ds.resize((dim1,))
                    ds[dim1-1] = data
                else:                               
                    ## list of entries               
                    if reset:
                        ## here the data gets not appended but overwritten!
                        ds.resize((len(data),))
                        ds[:] = data
                    else:
                        ## data append
                        dim1 = ds.shape[0] 
                        ds.resize((dim1+len(data),))
                        ds[dim1:] = data

        if len(ds.shape) == 2:       
            ## 2 dim dataset: matrix
            ## multiple inputs: list/np.array with one or multiple entries
            fill = ds.attrs.get('fill')
            dim1 = ds.shape[1]
            if len(data) == 1 and pointwise:
                dim0 = max(1, ds.shape[0])
                ## single entry; sorting like in the 'len(ds.shape) == 3' case
                if next_matrix:
                    dim0 += 1
                    fill[0] += 1
                    fill[1] = 0
                if dim0 == 1: # very first slice
                    fill[0] = 1
                    dim1 += 1
                ds.resize((dim0,dim1))
                fill[1] += 1
                ds[fill[0]-1,fill[1]-1] = data
            else: 
                ## list of entries, sort the data 'slice by slice'
                dim0 = ds.shape[0]
                fill[1] = len(data)
                if reset:
                    ds[dim0-1,:] = data  # reset overwrites last data series (last row matrix)
                else:  # standard reset = False
                    fill[0] += 1
                    ds.resize((dim0+1,len(data)))
                    ds[dim0,:] = data
            ds.attrs.modify('fill', fill)

        if len(ds.shape) == 3:      
            ## 3 dim dataset: box
            ## input: np.array with multiple entries
            dim0 = max(1, ds.shape[0])
            dim1 = ds.shape[1]
            fill = ds.attrs.get('fill')
            if next_matrix:
                dim0 += 1
                fill[0] += 1
                fill[1] = 0
            if dim0 == 1:
                fill[0] = 1
                dim1 += 1
            if not reset:  # standard reset = False
                ds.resize((dim0,dim1,len(data))) # Update array size
                fill[1] += 1 # Update write position
            ds[fill[0]-1,fill[1]-1] = data # Our indices start with one.
            ds.attrs.modify("fill", fill)

        self.flush()
        
    def flush(self):
        self.hf.flush()
        
    def close_file(self):
        # delegate close
        if self.newfile:
            self.entry.attrs["updating"] = False
        self.hf.close()
        
    def __getitem__(self,s):
        return self.hf[s]



import logging

from qkit import visa

from qkit.drivers.AbstractMicrowaveSource import AbstractMicrowaveSource


class RS_SGMA(AbstractMicrowaveSource):

    def __init__(self, name, address, reset=False):
        logging.info(__name__ + ' : Initializing instrument R&S SGS100A Microwave Source')
        super().__init__(name, tags=['physical'])

        self._address = address
        self._visainstrument = visa.instrument(self._address)
        if (reset):
            self.reset()
        else:
            self.get_all()

    def reset(self):
        logging.info(__name__ + ' : resetting instrument')
        self._visainstrument.write('*RST')
        self.get_all()

    def do_get_power(self):
        logging.debug(__name__ + ' : get power')
        return float(self._visainstrument.query('POW:POW?'))

    def do_set_power(self, amp):
        logging.debug(__name__ + ' : set power to %f' % amp)
        self._visainstrument.write('POW:POW %s' % amp)

    def do_get_frequency(self):
        logging.debug(__name__ + ' : get frequency')
        return float(self._visainstrument.query('FREQ:CW?'))

    def do_set_frequency(self, freq):
        logging.debug(__name__ + ' : set frequency to %f' % freq)
        self._visainstrument.write('FREQ:CW %s' % freq)

    def do_get_status(self):
        logging.debug(__name__ + ' : get status')
        return bool(int(self._visainstrument.query('OUTP?')))

    def do_set_status(self, status):
        logging.debug(__name__ + ' : set status to %s' % status)

        if status == True:
            self._visainstrument.write('OUTP ON')
        elif status == False:
            self._visainstrument.write('OUTP OFF')
        else:
            raise ValueError('set_status(): can only set True or False')



# -*- coding: utf-8 -*-
"""

@author: h.rotzinger
"""
from qkit.storage import hdf_lib
import numpy as np
from scipy.ndimage.filters import sobel
from scipy.ndimage import gaussian_filter1d
from scipy.ndimage import median_filter


class qps_ring(object):
    def __init__(self, file_path):
        self._hf = hdf_lib.Data(path=file_path)
        self._prepare_datasets()
        
    def _prepare_datasets(self):
        '''
        reads out the file
        '''    
        # these ds_url should always be present in a resonator measurement
        ds_url_amp = "/entry/data0/amplitude"
        ds_url_pha = "/entry/data0/phase"
        ds_url_freq = "/entry/data0/frequency"
    
    
        self._ds_amp = self._hf.get_dataset(ds_url_amp)
        self._ds_pha = self._hf.get_dataset(ds_url_pha)
    
        self._amplitude = np.array(self._hf[ds_url_amp],dtype=np.float64)
        self._phase = np.array(self._hf[ds_url_pha],dtype=np.float64)
        self._frequency = np.array(self._hf[ds_url_freq],dtype=np.float64)
    

        self._x_co = self._hf.get_dataset(self._ds_amp.x_ds_url)
 
        self._prepared_datasets = True

    def _prepare_find_jumps(self):
        self.qps_jpn_nr    = self._hf.add_value_vector('qps_jpn_nr', folder = 'analysis', 
                                                    x = self._x_co, unit = 'Num')
        self.qps_jpn_hight = self._hf.add_value_vector('qps_jpn_hight', folder = 'analysis', 
                                                    x = self._x_co, unit = 'Hz')
        self.qps_jpn_spec  = self._hf.add_value_vector('qps_jpn_spec', folder = 'analysis', 
                                                    x = self._x_co, unit = 'Hz')

    def find_jumps(self,ds, threshold = 40000):
        self._prepare_find_jumps()
        ds = self._hf[ds]
        ds = gaussian_filter1d(ds,2)
        offset=ds[0]
        jpnh = 0
        for i in xrange(ds.shape[0]-3):
            #i +=3
            #df=(((ds[i+1]+ds[i+2]+ds[i+3])/3.)-ds[i])
            #df=(ds[i] - ((ds[i-1]+ds[i-2]+ds[i-3])/3.))
            df=((ds[i+1])-ds[i])
            if (abs(df)>threshold):
                self.qps_jpn_nr.append(1.)
                offset = offset-df
                jpnh = df
                #print df, offset
                self.qps_jpn_hight.append(abs(float(jpnh)))
                
                self.qps_jpn_spec.append(float(ds[i]+offset))
                jpnh = df
            
            else:
                self.qps_jpn_nr.append(0.)
                #self.qps_jpn_hight.append(float(jpnh))
                self.qps_jpn_spec.append(float(ds[i]+offset))
    def find_jumps2(self,ds,threshold=30000):
        self._prepare_find_jumps()
        ds = self._hf[ds]
        offset=ds[0]
        # first we remove a bit of noise
        #flt = gaussian_filter1d(ds,10)
        flt = median_filter(ds,size=10)
        #flt = ds
        # the sobel filter finds the "jumps" 
        sb=sobel(flt)
        for i in sb:
            self.qps_jpn_hight.append(float(i))
            
        for i in flt: self.qps_jpn_spec.append(float(i))
        """    
        for i in xrange(flt.shape[0]-1):
            if(abs(sb[i])>threshold):
                offset -= sb[i]
                
                self.qps_jpn_spec.append(float(flt[i]-offset))
            else:
                self.qps_jpn_spec.append(float(flt[i]-offset))
        """       

        #for i in sb
        
    def split_traces(self,ds,threshold=30000):
        self._prepare_find_jumps()
        ds = self._hf[ds]
        # first we remove a bit of noise, size is the number of averages
        #flt = gaussian_filter1d(ds,10)
        flt = median_filter(ds,size=3)
        #flt = ds
        # the sobel filter finds the "jumps" 
        sb=sobel(flt)
        for i in sb:
            self.qps_jpn_hight.append(float(i))
            
        #for i in flt: self.qps_jpn_spec.append(float(i))
        offset=ds[0]
        tr_num = 0
        tr_name = "qps_tr_"+str(tr_num)
        tr_obj =  self._hf.add_value_vector(tr_name, 
                                            folder = 'analysis', 
                                            x = self._x_co, 
                                            unit = 'Hz')
        keepout = 4
        for i,tr in enumerate(flt):
            keepout += 1
            if abs(sb[i])>threshold and keepout>3:
                keepout = 0
                # new trace
                tr_num +=1
                tr_name = "qps_tr_"+str(tr_num)
                tr_obj =  self._hf.add_value_vector(tr_name, 
                                                    folder = 'analysis', 
                                                    x =  self._x_co, 
                                                    unit = 'Hz')
                print tr , i
                #tr_obj.append(float(tr))
            else:
                if keepout>2:
                    tr_obj.append(float(tr-offset))
        
                
            
if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(
        description="qps_rings: simple ring trace split/analysis/ KIT 2016 \n The traces have to be fitted first. Assumed is a skewed lorenzian fit.")

    parser.add_argument('-f','--file',     type=str, help='hdf filename to open')
    parser.add_argument('-tr','--threshold', type=float, help='(optional) treshold frequency for splitting the traces')
    #parser.add_argument('-fit','--fit-dataset',     type=str, help='dataset to split')

    args=parser.parse_args()
    if not args.file:
        print "please provide a hdf filename!"
        exit(1)
    qr = qps_ring(args.file)
    if args.threshold:
        qr.split_traces('/entry/analysis0/sklr_f0', threshold = args.threshold)
    else:
        qr.split_traces('/entry/analysis0/sklr_f0')


# HP_81110A.py class, to perform the communication between the Wrapper and the device
# Pieter de Groot <pieterdegroot@gmail.com>, 2008
# Martijn Schaafsma <qtlab@mcschaafsma.nl>, 2008
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from qkit.core.instrument_base import Instrument
from qkit import visa
import types
import logging
import time

class HP_81110A(Instrument):
    '''
    This is the python driver for the HP 81110A
    pulse generator
    Also works with the Agilent 81130A, the former HP 81130A.

    Usage:
    Initialize with
    <name> = instruments.create('<name>', 'HP_81110A', address='<GPIB address>',
        reset=<bool>)
    '''

    def __init__(self, name, address, reset=False):
        '''
        Initializes the HP_81110A, and communicates with the wrapper.

        Input:
            name (string)    : name of the instrument
            address (string) : GPIB address
            reset (bool)     : resets to default values, default=false

        Output:
            None
        '''

        Instrument.__init__(self, name, tags=['physical'])

        self._address = address
        self._visainstrument = visa.instrument(self._address)
        self._channels = self._get_number_of_channels()

        self.add_parameter('delay', type=float,
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            channels=(1, self._channels), minval=0.0, maxval=999, units='sec',channel_prefix='ch%d_')
        self.add_parameter('width', type=float,
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            channels=(1, self._channels), minval=-6.25e-9, maxval=999.5, units='sec',channel_prefix='ch%d_')
        self.add_parameter('high', type=float,
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            channels=(1, self._channels), minval=-9.90, maxval=10.0, units='Volts',channel_prefix='ch%d_')
        self.add_parameter('low', type=float,
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            channels=(1, self._channels), minval=-10.0, maxval=9.90, units='Volts',channel_prefix='ch%d_')
        self.add_parameter('status', type=str, channels=(1, self._channels),
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,channel_prefix='ch%d_')
        self.add_parameter('add_channels', type=str,
            flags=Instrument.FLAG_GETSET)# | Instrument.FLAG_GET_AFTER_SET)
        self.add_parameter('display', type=str,
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET)
        self.add_parameter('frequency', type=float, 
            flags=Instrument.FLAG_GETSET, units = 'Hz')

        self.add_function('reset')
        self.add_function('get_all')
        self.add_function('set_mode_triggered')
        self.add_function('set_mode_continuous')
        self.add_function('set_mode_gated')

        if reset:
            self.reset()
        else:
            self.get_all()



    def reset(self):
        '''
        Resets the instrument to default values

        Input:
            None

        Output:
            None
        '''
        logging.info(__name__ + ' : Resetting instrument')
        self._visainstrument.write('*RST')

        """ Fix: The Agilent 81130A is kinda slow after a reset. """
        time.sleep(2)

        self.get_all()

    def get_all(self):
        '''
        Reads all implemented parameters from the instrument,
        and updates the wrapper.

        Input:
            None

        Output:
            None
        '''
        logging.info(__name__ + ' : reading all settings from instrument')

        for i in range(1,self._channels+1):
            self.get('ch%d_delay' % i)
            self.get('ch%d_width' % i)
            self.get('ch%d_low' % i)
            self.get('ch%d_high' % i)
            self.get('ch%d_status' % i)

        self.get_display()
        self.get_frequency()

    # communication with device
    def do_get_frequency(self):
    	'''
    	Gets the pulse repetition frequency
    	
    	Output:
    		frequency (float): pulse repetition frequency in Hz
    	'''
    	return float(self._visainstrument.ask(':FREQ?'))
		
    def do_set_frequency(self, frequency):
    	'''
    	Sets the pulse repetition frequency
    	
    	Input:
    		frequency (float): pulse repetition frequency in Hz
    	'''
    	self._visainstrument.write(':FREQ %e'%frequency)

    def do_get_delay(self, channel):
        '''
        Reads the pulse delay from the specified channel

        Input:
            channel (int) : 1 or 2, the number of the designated channel

        Output:
            delay (float) : delay in seconds
        '''
        logging.debug(__name__ + ' : get delay for channel %d' % channel)
        return float(self._visainstrument.ask(':PULS:DEL' + str(channel) + "?"))

    def do_set_delay(self, val, channel):
        '''
        Sets the delay of the pulse of the specified channel

        Input:
            val (float)   : delay in seconds
            channel (int) : 1 or 2, the number of the designated channel

        Output:
            None
       
        self._visainstrument.write(':PULS:DEL' + str(channel) + " " + str(val) + "S")

    def do_get_width(self, channel):
        '''
        Reads the pulse width from the device for the specified channel

        Input:
            channel (int) : 1 or 2, the number of the designated channel

        Output:
            width (float) : width in seconds
        '''
        logging.debug(__name__ + ' : get width for channel %d' % channel)
        return float(self._visainstrument.ask(':PULS:WIDT' + str(channel) + "?"))

    def do_set_width(self, val, channel):
        '''
        Sets the width of the pulse of the specified channel

        Input:
            val (float)   : width in seconds
            channel (int) : 1 or 2, the number of the designated channel

        Output:
            None
        '''
        logging.debug(__name__ + ' : set width for channel %d to %f' % (channel, val))
        self._visainstrument.write(':PULS:WIDT' + str(channel) + " " + str(val) + "S")

    def do_get_high(self, channel):
        '''
        Reads the upper value from the device for the specified channel

        Input:
            channel (int) : 1 or 2, the number of the designated channel

        Output:
            val (float) : upper bound in Volts
        '''
        logging.debug(__name__ + ' : get high for channel %d' % channel)
        return float(self._visainstrument.ask(':VOLT' + str(channel) + ':HIGH?'))

    def do_set_high(self, val, channel):
        '''
        Sets the upper value of the specified channel

        Input:
            val (float)   : high bound in Volts
            channel (int) : 1 or 2, the number of the designated channel

        Output:
            None
        '''
        logging.debug(__name__ + ' : set high for channel %d to %f' % (channel, val))
        self._visainstrument.write(':VOLT' + str(channel) + ":HIGH " + str(val) + "V")

    def do_get_low(self, channel):
        '''
        Reads the lower value from the device for the specified channel

        Input:
            channel (int) : 1 or 2, the number of the designated channel

        Output:
            val (float) : lower bound in Volts
        '''
        logging.debug(__name__ + ' : get low for channel %d' % channel)
        return float(self._visainstrument.ask(':VOLT' + str(channel) + ':LOW?'))

    def do_set_low(self, val, channel):
        '''
        Sets the lower value of the specified channel

        Input:
            val (float)   : low bound in Volts
            channel (int) : 1 or 2, the number of the designated channel

        Output:
            None
        '''
        logging.debug(__name__ + ' : set low for channel %d to %f' % (channel, val))
        self._visainstrument.write(':VOLT' + str(channel) + ":LOW " + str(val)        + "V")

    def do_get_status(self, channel):
        '''
        Reads the status from the device for the specified channel

        Input:
            channel (int) : 1 or 2, the number of the designated channel

        Output:
            status (string) : 'on' or 'off'
        '''
        logging.debug(__name__ + ' : getting status for channel %d' % channel)
        val = self._visainstrument.ask('OUTP' + str(channel) + '?')
        if (val=='1'):
            return 'on'
        elif (val=='0'):
            return 'off'
        return 'error'

    def do_set_status(self, val, channel):
        '''
        Sets the status of the specified channel

        Input:
            val (string)  : 'on' or 'off'
            channel (int) : 1 or 2, the number of the designated channel

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting status for channel %d to %s' % (channel, val))
        if ((val.upper()=='ON') | (val.upper()=='OFF')):
            self._visainstrument.write('OUTP' + str(channel) + " " + val)
        else:
            logging.error('Try tot set OUTP to ' + str(val))

    def do_get_display(self):
        '''
        Reads the display status from the device

        Input:
            None

        Output:
            status (string) : 'on' or 'off'
        '''
        logging.debug(__name__ + ' : getting display status')
        val = self._visainstrument.ask('DISP?')
        if (val=='1'):
            return 'on'
        elif (val=='0'):
            return 'off'
        return 'error'

    def do_set_add_channels(self, val):
        '''
        Digitally add the pulses of the two channels

        Input:
            val (string)  : 'on' or 'off'

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting channel addition to %s' % (val))
        if (val.upper()=='ON'):
            self._visainstrument.write('CHAN:MATH DIG')
        elif (val.upper()=='OFF'):
            self._visainstrument.write('CHAN:MATH OFF')
        else:
            logging.error('Try tot set MATH to ' + str(val))

    def do_get_add_channels(self):
        '''
        Digitally add the pulses of the two channels

        Input:
            None

        Output:
            status (string) : 'on' or 'off'
        '''
        logging.debug(__name__ + ' : getting display status')
        val = self._visainstrument.ask('CHAN:MATH?')
        if (val=='DIG'):  return 'on'
        elif (val=='OFF'): return 'off'
        return 'error'

    def do_set_display(self, val):
        '''
        Sets the display status of the device

        Input:
            val (string) : 'on' or 'off'

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting display status to %s' % val)
        if ((val.upper()=='ON') | (val.upper()=='OFF')):
            self._visainstrument.write('DISP ' + val)
        else:
            logging.error('Try to set display to ' +val)

    def set_mode_triggered(self):
        '''
        Sets the instrument in 'triggered' mode

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting instrument to triggered mode')
        self._visainstrument.write(':ARM:SOUR EXT')

    def set_mode_continuous(self):
        '''
        Sets the instrument in 'continuous' mode

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting instrument to continuous mode')
        self._visainstrument.write(':ARM:SOUR IMM')

    def set_mode_gated(self):
        '''
        Sets the instrument in 'gated' mode

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting instrument to gated mode')
        self._visainstrument.write(':ARM:MODE GAT')


    def _get_number_of_channels(self):
        '''
        asks the device for the options installed and derives the number of channels
        Fixme: maybe there is a direct method to examine the number of channels.

        Input:
            None
        Output:
            Number of installed channels (int)

        '''
        print "PULSER-Notice: Number of channels is hardcoded to 1 for a quick fix"
        return 1
        opt = self._visainstrument.ask('*OPT?').split()
        return int(len(opt)-opt.count(0))



# -*- coding: utf-8 -*-
"""
Data point tracker by YS @ KIT / 2017
"""

import peakutils
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm


class pointtracker():
    """
    New point tracking routine, successor to the soon deprecated ac_preparation.
    Used to detect peaks or dips along branches (e.g. of anticrossings) in a 2d spectrum.
    
    Functions:
        set_data             -- defining the data set to be handled
        set_searchparams     -- setting the parameters for the search: start, span and point type
        set_peakutils_params -- if necessary, change the parameters of the used peakutils
        start_tracking       -- starting the actual point tracking routine
        del_trace            -- remove one of the detected point traces
        cut                  -- remove a given amount of detected points from a given trace
        del_points           -- remove a specific set of points from a given trace
        plot                 -- plot the data with an overlay of the detected points
        get_results          -- return the results in dimensions of xdata and ydata
    """
    
    def __init__(self):
        # initialize arrays
        self.x_results = []
        self.y_results = []
        
        # set default parameters for peakutils
        self._thres = 0.3
        self._min_dist = 1
   
    
    def set_data(self, xdata, ydata, data):
        """
        Defines the data arrays to be used by the pointtracker.
        Requires 3 arrays of a 2d spectrum
        
        Keyword arguments:
            xdata -- array of the sweep coordinates (e.g. current values)
            ydata -- usually the frequency array
            data  -- usually amplitude data, in the form of [len(xdata),len(ydata)]
        """
        
        self.xdata = xdata
        self.ydata = ydata
        self.data = data
        
        
    def set_searchparams(self, start_coords, span, dips=True):
        """
        Set the parameters for the search: start, span and point type
        Requires set_data to be run before
        
        Keyword arguments:
            start_coords -- guessed arbitrary point on branch in form [x,y]
                            and units of xdata and ydata
            span         -- the span along the y-axis in units of ydata
            dips         -- search for dips instead of peaks (default True)
        """
        
        try: self.data
        except AttributeError:
            print("Please run set_data first")
            return
        
        # if dips are searched invert the dataset
        if dips: self._sig = -1
        else: self._sig = +1
        
        # find closest real data point to given start_coords and translate in indices
        self.start_coords = []
        self.start_coords.append(np.argmin(np.abs(self.xdata-start_coords[0])))
        self.start_coords.append(np.argmin(np.abs(self.ydata-start_coords[1])))
        
        # translate given span in span of indices
        self.span=int(span/((self.ydata[-1]-self.ydata[0])/len(self.ydata)))
        
        
    def set_peakutils_params(self, thres=0.3, min_dist=1):
        """
        If necessary, change the parameters of the used peakutils
            
        Keyword arguments:
            thres (float)  -- normalized threshold (default: 0.3)
                              only peaks with amp higher than this will be detected
            min_dist (int) -- minimum distance between each detected peak (default: 1)
        """
        
        self._thres = thres
        self._min_dist = min_dist
        
    
    def start_tracking(self):
        """
        Start the actual point tracking routine.
        Requires set_data and set_searchparams to be run before.
        The points found along one branch are sorted and stored as numpy arrays of their
        x- and ydata indices which are appended to the x_results and y_results lists.
        """
        
        # initialize temporary point arrays
        self._points_x = np.array([],dtype="int")
        self._points_y = np.array([],dtype="int")
        
        # recursively find points
        self._track_points(self.start_coords, direction=0)
        
        # sort point arrays
        self._points_y = self._points_y[np.argsort(self._points_x)]
        self._points_x = np.sort(self._points_x)
        
        # append results to result arrays
        self.x_results.append(self._points_x)
        self.y_results.append(self._points_y)
        
        # clean up
        self._points_x = np.array([],dtype="int")
        self._points_y = np.array([],dtype="int")

    
    def _track_points(self, search_indices, direction):
        """
        The actual point tracking routine.
        Should only be run via the start_tracking function.
        Calls itself recursively and walks along one branch in both directions,
        starting from given start_coords and detects peaks or dips in a given span.
        For the next call, the search_indices are shifted depending on the distance
        between the last found points.
        Recursion stops as soon as the boundary of the dataset is reached in both directions.
        Detected points are added to the hidden _points_x and _points_y arrays.
        """
        
        # Test if still in data
        if (search_indices[0]<0 or search_indices[0]>=len(self.xdata) or
            search_indices[1]-self.span/2<0 or search_indices[1]+self.span/2>=len(self.ydata)):
            print("Reached boundary of dataset")
            return
        
        search_data = self._sig * self.data[int(search_indices[0]), int(search_indices[1]-self.span/2) : int(search_indices[1]+self.span/2)]

        indexes_found = peakutils.indexes(search_data, thres=self._thres, min_dist=self._min_dist)

        # add found peaks to arrays, and repeat recursively with shifted search_indices
        if indexes_found.size>=1:
            y_new = search_indices[1] - int(self.span/2) + indexes_found[0]
            self._points_y = np.append(self._points_y,y_new)
            self._points_x = np.append(self._points_x,search_indices[0])
            
            # If more than one peak found, print it and take the first one
            if indexes_found.size>1:
                print("Found more than one peak in trace " + str(search_indices[0]))
                print("First found peak added")
                
        # If no peak found, print it and continue with next current trace
        else:
            print("No peak found in trace " + str(search_indices[0]))
            # Add distance between two last found peaks to shift
            if len(self._points_y) >= 2:
                y_new = self._points_y[-1] + (self._points_y[-1] - self._points_y[-2])
            else:
                y_new = search_indices[1]
             
        # shift search intervall (if not the first point) and search in shifted intervall
        if direction==0:
            self._track_points([search_indices[0]+1,search_indices[1]], direction=1)
            self._track_points([search_indices[0]-1,search_indices[1]], direction=-1)
        else:
            search_indices[1] = search_indices[1] + (y_new - search_indices[1])
            self._track_points([search_indices[0]+direction,search_indices[1]], direction=direction)
            
            
    def del_trace(self, trace=-1, all=False):
        """
        Remove one of the detected point traces (default: last one).
        
        Keyword arguments:
            trace (int) -- # of trace to remove from results
            all (bool)  -- clear results
        """
        
        if all:
            for i in range(0,len(self.x_results)):
                self.x_results.pop(-1)
                self.y_results.pop(-1)
        else:
            self.x_results.pop(trace)
            self.y_results.pop(trace)
            
            
    def cut(self, amount=0, trace=-1, end="high"):
        """
        Remove a given amount (default: 0) of detected points from a given trace (default: -1).
        It starts from the given end (default: high end in x dimensions).
        
        Keyword arguments:
            amount (int) -- amount of points to be removed from the end of the trace
            trace (int)  -- # of trace to remove points from
            end (str)    -- from which end to start ("high" or "low")
        """
        
        if amount == 0:
            n = None
        else:
            if end == "high":
                n = -amount
                self.x_results[trace] = self.x_results[trace][0:n]
                self.y_results[trace] = self.y_results[trace][0:n] 
            else:
                n = amount
                self.x_results[trace] = self.x_results[trace][n:None]
                self.y_results[trace] = self.y_results[trace][n:None]        


    def del_points(self, indices=[], trace=-1):
        """
        Remove a specific set of points (default : []) from a given trace (default: -1).
        
        Keyword arguments:
            indices ([int]) -- list of indices of points to be removed from the trace
            trace (int)     -- # of trace to remove points from
        """
        
        self.x_results[trace] = np.delete(self.x_results[trace], indices)
        self.y_results[trace] = np.delete(self.y_results[trace], indices)
        
        
    def plot(self, all=True, amount=1, log=False):
        """
        Plot the data with an overlay of the detected points.
        
        Keyword arguments:
            all (bool)   -- plot all detected traces (default: True)
            amount (int) -- plot given amount of traces (default: 1)
            log (bool)   -- plot data logarithmically (default: False)
        """
        
        fig, axes = plt.subplots(figsize=(16,8))
        
        
        if log==False:
            plt.pcolormesh(self.xdata, self.ydata, self.data.T, cmap="viridis")
        else:
            plt.pcolormesh(self.xdata, self.ydata, self.data.T, cmap="viridis", norm=LogNorm(vmin=self.data.min(), vmax=self.data.max()))
        plt.xlim(min(self.xdata), max(self.xdata))
        plt.ylim(min(self.ydata), max(self.ydata))
        plt.colorbar()
        
        if all:
            n = len(self.x_results)
        else:
            n = amount
        col = ["r", "w", "m", "k", "b", "g", "c", "y"]
        if n>len(col):
            m=int(n/len(col)+1)
            col=m*col
        
        for i in range(0,n):
            plt.plot(self.xdata[self.x_results[i]], self.ydata[self.y_results[i]], col[i]+"x", label="Trace %d"%(i))
        
        plt.legend()
        
        
    def get_results(self):
        """
        Return the results in dimensions of xdata and ydata.
        Returns a list of two lists (for x and y) of numpy arrays (with the detected points).
        """
        
        xres = []
        yres = []
        
        for i in range(0,len(self.x_results)):
            xres.append(self.xdata[self.x_results[i]])
            yres.append(self.ydata[self.y_results[i]])
        
        return [xres,yres]


# DATA class v1.0 written by HR,JB@KIT 2011, 2016

'''
DATA exchange class, also holds global variables for thread management. 
Generalized version based on TIP.
'''


# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


import time
import os
from threading import Lock
import numpy as np

from qkit.storage import store as hdf_lib
from qkit.gui.plot import plot as qviewkit

class DATA(object):
    '''
    DATA class. Controls all access to parameter values and stored data.
    '''
    class LOCALHOST(object):
        def __init__(self,config):
            self.name = config.get('LOCALHOST','name')
            self.ip   = config.get('LOCALHOST','ip')
            self.port = config.getint('LOCALHOST','port')
    class REMOTEHOST(object):
        def __init__(self,config):
            self.name = config.get('REMOTEHOST','name')
            self.ip   = config.get('REMOTEHOST','ip')
            self.port = config.getint('REMOTEHOST','port')
            
    class PARAMETER(object):
        def __init__(self,config,p_index,p_attr):
            '''
            Initialize parameter attributes, mostly taken from the config file.
            '''
            self.p_index = p_index
            self.name = config.get(str(p_attr),'name')
            self.interval = config.getfloat(str(p_attr),'interval')
            self.data_request_object = lambda: 0   #this is written by server_main.py
            self.value = 0
            self.timestamp = 0
            self.next_schedule = time.time() + self.interval
            self.logging = bool(int(config.get(str(p_attr),'logging')))
            self.log_path = config.get(str(p_attr),'log_path')
            self.log_lock = Lock()
            self.url_timestamps = None
            print("Parameter %s loaded."%str(self.name))
            
        def get_all(self):
            '''
            Get all parameter attributes.
            '''
            with Lock():
                return {
                    "parameter": self.p_index,
                    "name":self.name,
                    "interval":self.interval,
                    "value":self.value,
                    "timestamp":self.timestamp,
                    "next_schedule":self.next_schedule,
                    }

        def set_interval(self,interval):
            '''
            Setup the scheduling, which corresponds to the value request interval.
            '''
            interval = float(interval)
            if interval == 0:
                interval = 120*365*24*3600 #120 years
            self.next_schedule += -self.interval+interval #update next schedule
            self.interval = interval   
        
        def get_last_value(self):
            with Lock():
                return self.value
        def get_timestamp(self):
            with Lock():
                return self.timestamp
        
        def get_history(self,range,nchunks=100):
            '''
            Read out the h5 file.
            - range: history data range
            - nchunks: number of data points to be returned
            '''
            
            if self.url_timestamps != None:
                try:
                    with self.log_lock:
                        timestamps = np.array(self.hf[self.url_timestamps])
                        values = np.array(self.hf[self.url_values])
                    data_points_requested_mask = np.where(time.time()-timestamps < range*3600)
                    timestamps_requested = timestamps[data_points_requested_mask]
                    data_points_requested = values[data_points_requested_mask]
                    #return only nchunks data points
                    if len(data_points_requested) > nchunks:
                        timestamp_chunks = np.array(np.split(timestamps_requested[:(len(timestamps_requested)-len(timestamps_requested)%nchunks)],nchunks))
                        timestamps = timestamp_chunks[:,int(0.5*len(timestamps_requested)/nchunks)]
                        data_chunks = np.array(np.split(data_points_requested[:(len(data_points_requested)-len(data_points_requested)%nchunks)],nchunks))
                        #calculate medians and return them instead of the mean (due to runaways in the log file)
                        medians = np.sort(data_chunks,axis=-1)[:,int(0.5*len(data_points_requested)/nchunks)]
                        return [timestamps,medians]
                    else:
                        return [timestamps_requested,data_points_requested]
                    
                except KeyError:   #AttributeError, NameError:
                    print('Error opening h5 log file.')
                    return [0]
            else:
                return [0]
        
        def store_value(self,value):
            '''
            Store method, used by the worker.
            '''
            with Lock():
                try:
                    self.value = float(value)
                except ValueError:
                    print('type cast error, ignoring')
                self.timestamp = time.time()
            if self.logging:
                self.append_to_log()
                
        def create_logfile(self):
            print('Create new log file for parameter %s.'%self.name)
            self.fname = os.path.join(self.log_path,self.name.replace(' ','_')+time.strftime('%d%m%Y%H%M%S')+'.h5')
            #print self.fname
            self.hf = hdf_lib.Data(self.fname, mode='a')
            self.hdf_t = self.hf.add_coordinate('timestamps')
            self.hdf_v = self.hf.add_value_vector('values', x = self.hdf_t)
            self.url_timestamps = '/entry/data0/timestamps'
            self.url_values = '/entry/data0/values'
            view = self.hf.add_view('data_vs_time', x = self.hdf_t, y = self.hdf_v)   #fit
        
        def append_to_log(self):
            with self.log_lock:
                self.hdf_t.append(float(self.get_timestamp()))
                self.hdf_v.append(float(self.get_last_value()))
            
        def close_logfile(self):
            self.hf.close_file()
        
        def schedule(self):
            '''
            Specifiy whether the parameter is to be updated, 
            typicalled called in each worker iteration. 
            Returns True if new parameter value needs to be read.
            '''
            if time.time() > self.next_schedule:
                while time.time() > self.next_schedule:
                    self.next_schedule += self.interval
                return True
            else:
                return False

        def set_schedule(self):
            self.next_schedule = time.time()
            return True
    
    def __init__(self,config):
        '''
        Reads the cfg file and instanciates all parameters accordingly.
        '''
        self.wants_abort = False
        self.debug = True
        self.cycle_time = config.getfloat('worker','cycle_time')
        
        p_instances = config.get('parameters','p').split(",")   #parameter instance names
        #print(p_instances)
        self.parameters = [self.PARAMETER(config,i,p) for i,p in enumerate(p_instances)]   #instanciate parameter array
        for i,p_i in enumerate(p_instances):   #create human readable aliases, such that objects are accessible from clients according to the seetings.cfg entry in []
            setattr(self,str(p_i),self.parameters[i])
        
        self.localhost = self.LOCALHOST(config)
        #self.remotehost = self.REMOTEHOST(config)
        self.ctrl_lock = Lock()

    def atexit(self):
        self.set_wants_abort()

    def get_wants_abort(self):
        with self.ctrl_lock:
            return self.wants_abort
    def set_wants_abort(self):
        with self.ctrl_lock:
            self.wants_abort = True
        

if __name__ == "__main__":
    DATA = DATA()



# brings QKIT around 
# HR@KIT/2017 based on work by R. Heeres/2008
# YS@KIT/2018: Further emancipated from qtlab
from __future__ import print_function

import qkit
import logging

from qkit.core.lib import temp

temp.File.set_temp_dir(qkit.cfg['tempdir'])

#
# assign instruments
#
from qkit.core.instrument_tools import Insttools

qkit.instruments = Insttools()

#
# assign flow
#
import qkit.core.flow as flow

qkit.flow = flow.FlowControl()
qkit.flow.sleep = qkit.flow.measurement_idle
qkit.flow.start = qkit.flow.measurement_start
qkit.flow.end = qkit.flow.measurement_end

if qkit.cfg.get("qt_compatible", False):
    raise ValueError("We do no longer provide legacy qtlab support. Please set qkit.cfg['qt_compatible']=False and clean up your code.")

# Other functions should be registered using qt.flow.register_exit_handler
from qkit.core.lib.misc import register_exit

register_exit(flow.qtlab_exit)



# qsurveilkit client written by HR,AS,JB@KIT 2013, 2016

'''
qsurveilkit client example, may run on any machine in the local network 
of the machine running server_main.py
'''

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


import sys
from threading import Thread

from PyQt4.QtCore import QObject, pyqtSignal
from PyQt4.QtGui import QMainWindow, QApplication

from plasma1_gui import Ui_MainWindow

import argparse
try:
    from ConfigParser import RawConfigParser
except ImportError:
    from configparser import RawConfigParser
import time
from math import log10

import numpy as np
from scipy.signal import medfilt as medfilter

import rpyc


class AcquisitionThread(Thread,QObject):
    """
    Acquisition loop. This is the worker thread that retrieves data from the server.
    """
    update_sig = pyqtSignal()
    
    def __init__(self,DATA):
        Thread.__init__(self)
        QObject.__init__(self)
        DATA.wants_abort = False
        self.data = DATA   #client data object
        
    def setup_remote_connection(self):
        '''
        Setup rpyc server connection.
        '''
        self.c = rpyc.connect(self.data.localhost.ip,self.data.localhost.port)
        print('Connection established to %s via port %s.'%(str(self.data.localhost.ip),str(self.data.localhost.port)))
        
    def acquire_datapoint_from_remote(self,p):
        '''
        Acquire the most recent data point of parameter instance p of the local client data object.
        '''
        return self.c.root.get_last_value(p.attribute_name)

    def acquire_history(self,p):
        '''
        acquires data from the log file of parameter p
        '''
        acq = self.c.root.get_history(p.attribute_name,p.range)
        if len(acq) == 2:
            return acq
        else:   #error reading h5 file
            raise IOError
    
    def close_connection(self):
        self.c.close()

    def run(self):

        self.setup_remote_connection()

        print('Client started.')
        while not self.data.wants_abort:
            for p in self.data.parameters:
                #extract new data point
                new_timestamp, new_value = self.acquire_datapoint_from_remote(p)
                p.values = np.append(p.values,new_value)
                p.timestamps = np.append(p.timestamps,new_timestamp)
            
                #check for data history request
                # if logging is active and more data is requested by the local GUI
                if p.logging and p.check_history and time.time() - p.timestamps[0] < p.range*3600:
                    try:
                        timestamps, values = self.acquire_history(p)
                        if timestamps[0] < p.timestamps[0]:
                            p.timestamps = np.array(timestamps, dtype='float64')
                            p.values = np.array(values, dtype='float32')
                    except IOError:
                        print('Error retrieving history data for parameter %s.'%str(p.attribute_name))
                p.check_history = False   #switch off history request after one unsuccessful attempt
                
                #delete old data points
                data_points_out_of_range = np.where(time.time()-p.timestamps > p.range*3600)
                p.values = np.delete(p.values,data_points_out_of_range)
                p.timestamps = np.delete(p.timestamps,data_points_out_of_range)
                
            self.update_sig.emit()   #tell the GUI that the client data set has been updated
            
            time.sleep(self.data.update_interval)
            
        self.close_connection()
        print('Connection closed.')


class MainWindow(QMainWindow, Ui_MainWindow):

    def myquit(self):
        exit()

    def __init__(self,DATA):
        
        self.data = DATA

        QMainWindow.__init__(self)
        # set up User Interface (widgets, layout...)
        self.setupUi(self)
        
        # --- specify graph and label objects as declared in the gui file
        self.data.pLL.graph = self.graph_LL
        self.data.pLL.label = self.label_LL
        self.data.pMC.graph = self.graph_MC
        self.data.pMC.label = self.label_MC
        
        #combo box entries, dictionary
        self.cb_scale_entries = {0:24,1:2,2:1,3:0.16}
        
        
        self.range = self._get_range()
        for p in self.data.parameters: p.range = self._get_range()
        
        self._setup_signal_slots()
        self._setup_graphs()
        
        self._start_acquisition()
        
        
    def _setup_signal_slots(self):
        self.cb_scale.currentIndexChanged.connect(self._range_changed)   #combo box
        
    def closeEvent(self,event):
        self.data.wants_abort = True
        print('Client closing ...')
        event.accept()
    
    def _setup_graphs(self):
        self.data.pLL.graph.setLabel('left',"LL pressure (mbar)")
        self.data.pLL.graph.setLabel('bottom',"time (h)")
        self.data.pLL.graph.plt = self.graph_LL.plot(pen='b')
        self.data.pLL.graph.setLogMode(y=True)
        
        self.data.pMC.graph.setLabel('left',"MC pressure (mbar)")
        self.data.pMC.graph.setLabel('bottom',"time (h)")
        self.data.pMC.graph.plt = self.graph_MC.plot(pen='b')
        self.data.pMC.graph.setLogMode(y=True)
        
    def _update_labels(self):
        self.data.pLL.label.setText('LL: %.3g mbar'%self.data.pLL.values[-1])
        self.data.pMC.label.setText('MC: %.3g mbar'%self.data.pMC.values[-1])
        
    def _autorange_plots(self):
        for p in self.data.parameters:
            p.graph.enableAutoRange()
        
    def _get_range(self):
        return self.cb_scale_entries[self.cb_scale.currentIndex()]
        
    def _range_changed(self):   #combo box change event
        if self.data.debug: print('range_changed')
        for p in self.data.parameters:
            p.range = self._get_range()
            p.check_history = True
        
    def _start_acquisition(self):
        print('Start acquisition.')
        self.acquisition_thread = AcquisitionThread(self.data)
        self.acquisition_thread.update_sig.connect(self._update_gui)
        self.acquisition_thread.start()
            
        
    @pyqtSlot(float)
    def _update_gui(self):
        for p in [self.data.pLL,self.data.pMC]:
            #remove error entries from values
            for i in range(1,len(p.values)-1):
                if len(p.values[i-4:i]) > 0 and len(p.values[i+1:i+5]) > 0:
                    median = np.sort(p.values[i-4:i+5])[int(0.5*len(p.values[i-4:i+5]))]
                    if np.abs(log10(p.values[i] / median)) > 6 and np.abs(log10(p.values[i] / p.values[i-1])) > 6:
                        #if jumps detected collecting more than 5 orders of magnitude back or forth
                        p.values[i] = median
                        if self.data.debug: print('runaway detected in {:s}, segment #{:d}'.format(p.attribute_name,i))
            p.graph.plt.setData(x=np.array((p.timestamps-time.time())/3600), y=p.values)
            self._autorange_plots()
            self._update_labels()

class DATA(object):
    '''
    Client data class. The __init__ method accesses the same cfg file as the server.
    '''
    class LOCALHOST(object):
        def __init__(self,config):
            self.name = config.get('LOCALHOST','name')
            self.ip   = config.get('LOCALHOST','ip')
            self.port = config.getint('LOCALHOST','port')
    class PARAMETER(object):
        '''
        Parameter class. Information is extracted from the cfg file.
        '''
        def __init__(self,config,p_index,p_attr):
            self.p_index = p_index
            self.attribute_name = str(p_attr)
            self.name = config.get(str(p_attr),'name')
            self.values = np.array([])
            self.timestamps = np.array([])
            self.logging = bool(int(config.get(str(p_attr),'logging')))
            self.check_history = False
        
    def __init__(self,config):
        self.debug = True
        self.wants_abort = False
        
        self.update_interval = config.getfloat('gui','update_interval')
        
        p_instances = config.get('parameters','p').split(",")   #parameter instance names
        #instanciate parameter objects
        self.parameters = [self.PARAMETER(config,i,p) for i,p in enumerate(p_instances)]   #instanciate parameter array
        for i,p_i in enumerate(p_instances):   #create human readable aliases, such that objects are accessible from clients according to the seetings.cfg entry in []
            setattr(self,str(p_i),self.parameters[i])
            
        self.localhost = self.LOCALHOST(config)
 
def main(argv):

    parser = argparse.ArgumentParser(
        description="Derived from TIP v0.1 (HR@KIT 2011), updated and generalized by HR,JB@KIT 2016")

    parser.add_argument('ConfigFile', nargs='?', default='settings.cfg',
                        help='Configuration file name')
    args=parser.parse_args()
    
    Conf = RawConfigParser()
    Conf.read(args.ConfigFile)
    
    data = DATA(Conf)
    data.config = Conf

    try:
        # create Qt application
        app = QApplication(argv,True)
        # create main window
        wnd = MainWindow(data) # classname
        wnd.show()
        
        # Connect signal for app finish
        app.connect(app, SIGNAL("lastWindowClosed()"), app, SLOT("quit()"))
        
        # Start the app up
        sys.exit(app.exec_())
    except KeyboardInterrupt:
        data.wants_abort = True
        print('Exiting client.')
        exit()
        
    
if __name__ == "__main__":
    main(sys.argv)


# -*- coding: utf-8 -*-
# IV_curve.py analysis class for IV like transport measurements
# Micha Wildermuth, micha.wildermuth@kit.edu 2019

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import sys
import numpy as np
from scipy import signal as sig
# TODO: uncertainty analysis probably using import uncertainties

import qkit
from qkit.storage.store import Data
import qkit.measure.measurement_class as mc
from qkit.gui.plot import plot as qviewkit

import json
from qkit.measure.json_handler import QkitJSONEncoder, QkitJSONDecoder

class dict2obj(object):
    def __init__(self, d):
        for a, b in d.items():
            if isinstance(b, (list, tuple)):
                setattr(self, a, [dict2obj(x) if isinstance(x, dict) else x for x in b])
            else:
                setattr(self, a, dict2obj(b) if isinstance(b, dict) else b)


class IV_curve2(object):
    """
    This is an analysis class IV like transport measurements taken by qkit.measure.transport.transport.py
    """

    def __init__(self):
        """
        Initializes an analysis class IV like transport measurements taken by qkit.measure.transport.transport.py
        Parameters
        ----------
        None
        Returns
        -------
        None
        Examples
        --------
        >>> import numpy as np
        >>> import qkit
        QKIT configuration initialized -> available as qkit.cfg[...]
        >>> qkit.start()
        Starting QKIT framework ... -> qkit.core.startup
        Loading module ... S10_logging.py
        Loading module ... S12_lockfile.py
        Loading module ... S14_setup_directories.py
        Loading module ... S20_check_for_updates.py
        Loading module ... S25_info_service.py
        Loading module ... S30_qkit_start.py
        Loading module ... S65_load_RI_service.py
        Loading module ... S70_load_visa.py
        Loading module ... S80_load_file_service.py
        Loading module ... S85_init_measurement.py
        Loading module ... S98_started.py
        Loading module ... S99_init_user.py
        Initialized the file info database (qkit.fid) in 0.000 seconds.
        >>> from qkit.analysis.IV_curve import IV_curve as IVC
        >>> ivc = IVC()
        Initialized the file info database (qkit.fid) in 0.000 seconds.
        """
        qkit.fid.update_file_db()  # update file database
        self.uuid, self.path, self.df = None, None, None
        self.settings = None
        self.mo = mc.Measurement()  # qkit-sample object
        self.m_type, self.scan_dim, self.sweeptype, self.sweeps, self.bias = None, None, None, None, None
        self.I, self.V, self.V_corr, self.dVdI = None, None, None, None
        self.I_offsets, self.V_offsets, self.I_offset, self.V_offset = None, None, None, None
        self.x_ds, self.x_coordname, self.x_unit, self.x_vec = None, None, None, None
        self.y_ds, self.y_coordname, self.y_unit, self.y_vec = None, None, None, None

    def load(self, uuid, dVdI='analysis0'):
        """
        Loads transport measurement data with given uuid <uuid>.

        Parameters
        ----------
        uuid: str
            qkit identification name, that is looked for and loaded
        dVdI: str | boolean
            Folder, where numerical derivative dV/dI is tried to load form datafile, if this was already analyzed during the measurement. If False, dV/dI is not loaded. Default is 'analysis0'.

        Returns
        -------
        None

        Examples
        --------
        >>> ivc.load(uuid='XXXXXX')
        """
        if uuid != self.uuid and self.uuid is not None:
            self.__init__()
        self.uuid = uuid
        self.path = qkit.fid.get(self.uuid)
        self.df = Data(self.path)
        try:
            self.settings = dict2obj(json.loads(self.df.data.settings.value[0], cls=QkitJSONDecoder))
        except:
            self.settings = self.df.data.settings[:]
        self.mo.load(qkit.fid.measure_db[self.uuid])
        self.m_type = self.mo.measurement_type  # measurement type
        if self.m_type == 'transport':
            self.scan_dim = self.df.data.i_0.attrs['ds_type']  # scan dimension (1D, 2D, ...)
            self.bias = self.get_bias()
            self.sweeps = self.mo.sample.sweeps  # sweeps (start, stop, step)
            self.sweeptype = self.get_sweeptype()
            shape = np.concatenate([[len(self.sweeps)], np.max([self.df['entry/data0/i_{:d}'.format(j)].shape for j in range(len(self.sweeps))], axis=0)])  # (number of sweeps, eventually len y-values, eventually len x-values, maximal number of sweep points)
            self.I, self.V, self.dVdI = np.empty(shape=shape), np.empty(shape=shape), np.empty(shape=shape)
            for j in range(shape[0]):
                i = self.df['entry/data0/i_{:d}'.format(j)][:]
                v = self.df['entry/data0/v_{:d}'.format(j)][:]
                if dVdI:
                    try:
                        dvdi = self.df['entry/{:s}/dvdi_{:d}'.format(dVdI, j)][:]  # if analysis already done during measurement
                    except KeyError:
                        dvdi = self.get_dydx(x=self.I[j], y=self.V[j])
                pad_width = np.insert(np.diff([i.shape, shape[1:]], axis=0), (0,), np.zeros(self.scan_dim)).reshape(self.scan_dim, 2)
                if pad_width.any():
                    self.I[j] = np.pad(i, pad_width, 'constant', constant_values=np.nan)  # fill with current values (eventually add nans at the end, if sweeps have different lengths)
                    self.V[j] = np.pad(v, pad_width, 'constant', constant_values=np.nan)  # fill with voltage values (eventually add nans at the end, if sweeps have different lengths)
                    if dVdI:
                        self.dVdI[j] = np.pad(dvdi, pad_width, 'constant', constant_values=np.nan)  # fill with differential resistance values (eventually add nans at the end, if sweeps have different lengths)
                else:
                    self.I[j] = i
                    self.V[j] = v
                    if dVdI:
                        self.dVdI[j] = dvdi
            if self.scan_dim >= 2:  # 2D or 3D scan
                # x parameter
                self.x_ds = self.df[self.df.data.i_0.attrs['x_ds_url']]
                self.x_coordname = self.x_ds.attrs['name']
                self.x_unit = self.x_ds.attrs['unit']
                self.x_vec = self.x_ds[:]
                if self.scan_dim == 3:  # 3D scan
                    # y parameter
                    self.y_ds = self.df[self.df.data.i_0.attrs['y_ds_url']]
                    self.y_coordname = self.y_ds.attrs['name']
                    self.y_unit = self.y_ds.attrs['unit']
                    self.y_vec = self.y_ds[:]
                else:
                    self.y_ds, self.y_coordname, self.y_unit, self.y_vec = None, None, None, None
            else:
                self.x_ds, self.x_coordname, self.x_unit, self.x_vec = None, None, None, None
        else:
            raise ValueError('No data of transport measurements')
        return

    def save(self, filename, params=None):
        """
        Saves the class variables
             * uuid,
             * path,
             * measurement_object,
             * measurement_type,
             * scan_dimension,
             * sweep_type,
             * sweeps,
             * bias,
             * I,
             * V,
             * V_corr (only if set
             * dVdI,
             * I_offsets,
             * V_offsets,
             * I_offset,
             * V_offset,
             * x_coordname,
             * x_unit,
             * x_vector,
             * y_coordname,
             * y_unit,
             * y_vector
        as well as <params> to a json-file.
        Parameters
        ----------
        filename: str
            Filename of the .json-file that is created.
        params: dict
            Additional variables that are saved. Default is None, so that only the above mentioned class variables are saved.
        Returns
        -------
        None
        """
        params = params if params else {}
        params = dict(
            {'uuid': self.uuid,
             'path': self.path,
             'measurement_object': self.mo.get_JSON(),
             'measurement_type': self.m_type,
             'scan_dimension': self.scan_dim,
             'sweep_type': self.sweeptype,
             'sweeps': self.sweeps,
             'bias': self.bias,
             'I': self.I,
             'V': self.V,
             'dVdI': self.dVdI,
             'I_offsets': self.I_offsets,
             'V_offsets': self.V_offsets,
             'I_offset': self.I_offset,
             'V_offset': self.V_offset,
             'x_coordname': self.x_coordname,
             'x_unit': self.x_unit,
             'x_vector': self.x_vec,
             'y_coordname': self.y_coordname,
             'y_unit': self.y_unit,
             'y_vector': self.y_vec}.items()
            + params.items()
        )
        if self.V_corr:
            params['V_corr'] = self.V_corr
        if '.json' not in filename:
            filename += '.json'
        with open(filename, 'w') as filehandler:
            json.dump(obj=params, fp=filehandler, indent=4, cls=QkitJSONEncoder, sort_keys=True)

    def open_qviewkit(self, uuid=None, ds=None):
        """

        uuid: str
        ds: str | list(str)
            Datasets that are opened instantaneously. Default is 'views/IV'
        """
        if uuid is None:
            df = self.df
        else:
            df = Data(qkit.fid.get(uuid))
        if ds is None:
            ds = ['views/IV']
            # try:
            #     if self.scan_dim > 1:
            #         for i in range(len(self.sweeps)):
            #             datasets.append('{:s}_{:d}'.format({0: 'I', 1: 'V'}[not self.bias].lower(), i))
        elif not np.iterable(ds) and type(ds) is str:
            ds = [ds]
        else:
            raise ValueError('Argument <ds> needs to be set properly.')
        qviewkit.plot(df.get_filepath(), datasets=ds)  # opens IV-view by default

    def get_bias(self, df=None):
        """
        Gets bias mode of the measurement. Evaluate 'x_ds_url' (1D), 'y_ds_url' (2D), 'z_ds_url' (3D) of i_0 and v_0 and checks congruence.
        Parameters
        ----------
        df: qkit.storage.store.Data (optional)
            Datafile of transport measurement. Default is None that means self.df
        Returns
        -------
        mode: int
            Bias mode. Meanings are 0 (current) and 1 (voltage).
        """
        #
        if df is None:
            df = self.df
        self.bias = {'i': 0, 'v': 1}[(
            df.data.i_0.attrs.get('{:s}_ds_url'.format(chr(self.scan_dim + 119))) and
            df.data.v_0.attrs.get('{:s}_ds_url'.format(chr(self.scan_dim + 119)))).split('/')[-1][0]]
        return self.bias

    def get_sweeptype(self, sweeps=None):
        """
        Gets the sweeptype of predefined set of sweeps as generated by qkit.measure.transport.transport.py
        Parameters
        ----------
        sweeps: array_likes of array_likes of floats (optional)
            Set of sweeps containing start, stop and step size (e.g. sweep object using qkit.measure.transport.transport.sweep class). Default is None that means self.sweeps.
        Returns
        -------
        sweeptype: int
            Type of set of sweeps. Is 0 (halfswing), 1 (4 quadrants) or None (arbitrary set of sweeps).
        """
        if sweeps is None:
            sweeps = self.sweeps
        # check if sweeps are halfswing
        if len(sweeps) == 2:
            if (np.array(sweeps[0])[[1, 0, 2]] == np.array(sweeps[1])[:3]).all():
                self.sweeptype = 0
        # check if sweeps are 4quadrants
        elif len(sweeps) == 4:
            if (np.array(sweeps[0])[[1, 0, 2]] == np.array(sweeps[1])[:3]).all() and (np.array(sweeps[2])[[1, 0, 2]] == np.array(sweeps[3])[:3]).all():
                self.sweeptype = 1
        else:
            self.sweeptype = None
        return self.sweeptype

    def get_dVdI(self, I=None, V=None, mode=sig.savgol_filter, **kwargs):
        """
        Calculates numerical derivative dV/dI.
        Parameters
        ----------
        I: numpy.array (optional)
            An N-dimensional array containing current values. Default is None that means self.I.
        V: numpy.array (optional)
            An N-dimensional array containing voltage values. Default is None that means self.V.
        mode: function (optional)
            Function that calculates the numerical gradient dx from a given array x. Default is scipy.signal.savgol_filter (Savitzky Golay filter).
        kwargs:
            Keyword arguments forwarded to the function <mode>.
        Returns
        -------
        dVdI: numpy.array
            Numerical derivative dV/dI
        Examples
        --------
        # Savitzky Golay filter
        >>> ivc.get_dVdI()
        # numerical gradient
        >>> ivc.get_dVdI(mode=np.gradient)
        """
        if V is None:
            y = self.V
        else:
            y = V
        if I is None:
            x = self.I
        else:
            x = I
        self.dVdI = self.get_dydx(y=y, x=x, mode=mode, **kwargs)
        return self.dVdI

    def get_dydx(self, y, x=None, mode=sig.savgol_filter, **kwargs):
        """
        Calculates numerical derivative dy/dx
        Parameters
        ----------
        y: numpy.array
            An N-dimensional array containing y-values
        x: numpy.array (optional)
            An N-dimensional array containing x-values. Default is None which means that x is considered as index.
        mode: function (optional)
            Function that calculates the numerical gradient dx from a given array x. Default is scipy.signal.savgol_filter (Savitzky Golay filter).
        kwargs:
            Keyword arguments forwarded to the function <mode>. Default for scipy.signal.savgol_filter is {'window_length': 15, 'polyorder': 3, 'deriv': 1} and for numpy.gradient {'axis': self.scan_dim}
        Returns
        -------
        dy/dx: numpy.array
            Numerical gradient quotient. If no x is given, dx = np.ones(nop)
        Examples
        --------
        Savitzky Golay filter
        >>> ivc.get_dydx(y=np.arange(1e1), mode=sig.savgol_filter, window_length=9, polyorder=3, deriv=1)
        array([1., 1., 1., 1., 1., 1., 1., 1., 1., 1.])
        numerical gradient
        >>> ivc.get_dydx(mode=np.gradient)
        ivc.get_dydx(y=np.arange(1e1), mode=np.gradient)
        """
        if mode == sig.savgol_filter:
            if 'window_length' not in kwargs.keys():
                kwargs['window_length'] = 15
            if 'polyorder' not in kwargs.keys():
                kwargs['polyorder'] = 3
            if 'deriv' not in kwargs.keys():
                kwargs['deriv'] = 1
        elif mode == np.gradient:
            if 'axis' not in kwargs.keys():
                kwargs['axis'] = self.scan_dim
        if x is None:
            if np.isnan(x).any():
                y_nans = np.isnan(y)
                dy = mode(np.nan_to_num(y, copy=True, nan=0.0), **kwargs)
                np.place(dy, y_nans, np.nan)
                return dy
            else:
                return mode(y, **kwargs)
        else:
            if np.isnan(x).any():
                x_nans = np.isnan(x)  # mask for np.nan
                y_nans = np.isnan(y)
                dx = mode(np.nan_to_num(x, copy=True, nan=0.0), **kwargs)  # derivation function with np.nan replaced by 0
                dy = mode(np.nan_to_num(y, copy=True, nan=0.0), **kwargs)
                np.place(dx, x_nans, np.nan)  # write np.nan using mask from above
                np.place(dy, y_nans, np.nan)
                return dy/dx
            else:
                return mode(y, **kwargs)/mode(x, **kwargs)

    def get_offsets(self, x=None, y=None, threshold=20e-6, offset=0, yr=False):
        """
        Calculates x- and y-offset for every trace. Therefore the branch where the y-values are nearly constant are evaluated. The average of all corresponding x-values is considered to be the x-offset and the average of the extreme y-values are considered as y-offset.
        Parameters
        ----------
        x: numpy.array (optional)
            An N-dimensional array containing x-values. Default is None, where x is considered as self.V and self.I in the current and voltage bias, respectively.
        y: numpy.array (optional)
            An N-dimensional array containing y-values. Default is None, where y is considered as self.I and self.V in the current and voltage bias, respectively.
        threshold: float (optional)
            Threshold voltage that limits the superconducting branch. Default is 20e-6.
        offset: float (optional)
            Voltage offset that shifts the limits of the superconducting branch which is set by <threshold>. Default is 0.
        yr: bool (optional)
            Condition, if critical or retrapping y-values are evaluated. Default is False.
        Returns
        -------
        I_offsets: numpy.array
            Current offsets of every single trace.
        V_offsets: numpy.array
            Voltage offsets of every single trace.
        """
        if x is None:
            x = [self.V, self.I][self.bias]
        if y is None:
            y = [self.I, self.V][self.bias]
        ''' constant range via threshold (for JJ superconducting range via threshold voltage) '''
        with np.errstate(invalid='ignore'):  # raises warning due to np.nan
            mask = np.logical_and(x >= -threshold + offset, x <= threshold + offset)
        x_const, y_const = np.copy(x), np.copy(y)
        np.place(x_const, np.logical_not(mask), np.nan)
        np.place(y_const, np.logical_not(mask), np.nan)
        if self.sweeptype == 0:  # halfswing
            ''' get x offset (for JJ voltage offset)'''
            x_offsets = np.mean(np.nanmean(x_const, axis=self.scan_dim), axis=0)
            ''' get y offset (for JJ current offset) '''
            if yr:  # retrapping y (for JJ retrapping current)
                y_rs = np.array([np.nanmax(y_const[0], axis=(self.scan_dim-1)),
                                 np.nanmin(y_const[1], axis=(self.scan_dim-1))])
                y_offsets = np.mean(y_rs, axis=0)
            else:  # critical y (for JJ critical current)
                y_cs = np.array([np.nanmin(y_const[0], axis=(self.scan_dim-1)),
                                 np.nanmax(y_const[1], axis=(self.scan_dim-1))])
                y_offsets = np.mean(y_cs, axis=0)
        elif self.sweeptype == 1:  # 4 quadrants
            ''' get x offset '''
            x_offsets = np.mean(np.nanmean(x_const, axis=self.scan_dim), axis=0)
            ''' get y offset '''
            if yr:  # retrapping y (for JJ retrapping current)
                y_rs = np.array([np.nanmax(y_const[1], axis=(self.scan_dim-1)),
                                 np.nanmin(y_const[3], axis=(self.scan_dim-1))])
                y_offsets = np.mean(y_rs, axis=0)
            else:  # critical y (for JJ critical current)
                y_cs = np.array([np.nanmax(y_const[0], axis=(self.scan_dim-1)),
                                 np.nanmin(y_const[2], axis=(self.scan_dim-1))])
                y_offsets = np.mean(y_cs, axis=0)
        else:  # custom sweeptype
            ''' get x offset '''
            x_offsets = np.nanmean(np.nanmean(x_const, axis=self.scan_dim), axis=0)
            ''' get y offset '''
            if yr:  # retrapping y (for JJ retrapping current)
                raise NotImplementedError('No algorithm implemented for custom sweeptype')
            else:
                y_cs = np.array([np.nanmax(y_const, axis=(self.scan_dim-1)),
                                 np.nanmin(y_const, axis=(self.scan_dim-1))])
                y_offsets = np.mean(y_cs, axis=0)
        self.I_offsets, self.V_offsets = [x_offsets, y_offsets][::int(np.sign(self.bias - .5))]
        return self.I_offsets, self.V_offsets

    def get_offset(self, *args, **kwargs):
        """
        Calculates x- and y-offset for the whole data set. Therefore the branch where the y-values are nearly constant are evaluated. The average of all corresponding x-values is considered to be the x-offset and the average of the extreme y-values are considered as y-offset.
        Parameters
        ----------
        x: numpy.array (optional)
            An N-dimensional array containing x-values. Default is None, where x is considered as self.V and self.I in the current and voltage bias, respectively.
        y: numpy.array (optional)
            An N-dimensional array containing y-values. Default is None, where y is considered as self.I and self.V in the current and voltage bias, respectively.
        threshold: float (optional)
            Threshold voltage that limits the superconducting branch. Default is 20e-6.
        offset: float (optional)
            Voltage offset that shifts the limits of the superconducting branch which is set by <threshold>. Default is 0.
        yr: bool (optional)
            Condition, if critical or retrapping y-values are evaluated. Default is False.
        Returns
        -------
        I_offsets: numpy.array
            Current offsets of the whole data set.
        V_offsets: numpy.array
            Voltage offsets of the whole data set.
        """
        self.I_offset, self.V_offset = map(np.nanmean, self.get_offsets(*args, **kwargs))
        return self.I_offset, self.V_offset

    def get_2wire_slope_correction(self, I=None, V=None, dVdI=None, peak_finder=sig.find_peaks, **kwargs):
        """
        Gets voltage values corrected by an ohmic slope such as occur in 2wire measurements.
        The two maxima in the differential resistivity <dVdI> are identified as critical and retrapping currents. The slope of the superconducting regime in between (which should ideally be infinity) is fitted using numpy.linalg.qr algorithm and subtracted from the raw data.
        Parameters
        ----------
        I: numpy.array (optional)
            An N-dimensional array containing current values. Default is None that means self.I.
        V: numpy.array (optional)
            An N-dimensional array containing voltage values. Default is None that means self.V.
        dVdI: numpy.array (optional)
            An N-dimensional array containing differential resistance (dV/dI) values. Default is None that means self.dVdI.
        peak_finder: function (optional)
            Peak finding algorithm. Default is scipy.signal.find_peaks.
        kwargs:
            Keyword arguments forwarded to the peak finding algorithm <peak_finder>.
        Returns
        -------
        V_corr: numpy.array
            Ohmic slope corrected voltage values
        """
        def lin_fit(x, y):
            X = np.stack((x, np.ones(len(x)))).T
            q, r = np.linalg.qr(X)
            p = np.dot(q.T, y)
            return np.dot(np.linalg.inv(r), p)
        if I is None:
            I = self.I
        if V is None:
            V = self.V
        if dVdI is None:
            dVdI = self.dVdI
        if peak_finder is sig.find_peaks and 'prominence' not in kwargs.keys():
            kwargs['prominence'] = 100
        if peak_finder is sig.find_peaks_cwt and 'widths' not in kwargs.keys():
            kwargs['widths'] = np.arange(10)
        ''' peak detection in dV/dI '''
        if self.scan_dim == 1:
            peaks = np.array(map(lambda dVdI1D:
                                 peak_finder(dVdI1D, **kwargs),
                                 dVdI))
            slices = map(lambda peaks1D:
                         slice(*np.sort(peaks1D[0][peaks1D[1]['prominences'].argsort()[-2:][::-1]])),
                         peaks)
            popts = map(lambda (I1D, V1D, s1D):
                        lin_fit(I1D[s1D], V1D[s1D]),
                        zip(I, V, slices))
            self.V_corr = np.array(map(lambda (I1D, V1D, popt1D):
                                       V1D - (popt1D[0] * I1D + popt1D[1]),
                                       zip(I, V, popts)))
            return self.V_corr
        elif self.scan_dim == 2:
            peaks = np.array(map(lambda dVdI2D:
                                 map(lambda dVdI1D: peak_finder(dVdI1D, **kwargs),
                                     dVdI2D),
                                 dVdI))
            slices = map(lambda peaks2D:
                         map(lambda peaks1D:
                             slice(*np.sort(peaks1D[0][peaks1D[1]['prominences'].argsort()[-2:][::-1]])),
                             peaks2D),
                         peaks)
            popts = map(lambda (I2D, V2D, s2D):
                        map(lambda (I1D, V1D, s1D):
                            lin_fit(I1D[s1D], V1D[s1D]),
                            zip(I2D, V2D, s2D)),
                        zip(I, V, slices))
            self.V_corr = np.array(map(lambda (I2D, V2D, popt2D):
                                       map(lambda (I1D, V1D, popt1D):
                                           V1D - (popt1D[0] * I1D + popt1D[1]),
                                           zip(I2D, V2D, popt2D)),
                                       zip(I, V, popts)))
            return self.V_corr
        elif self.scan_dim == 3:
            peaks = np.array(map(lambda dVdI3D:
                                 map(lambda dVdI2D:
                                     map(lambda dVdI1D:
                                         peak_finder(dVdI1D, **kwargs),
                                         dVdI2D),
                                     dVdI3D),
                                 dVdI))
            slices = map(lambda peaks3D:
                         map(lambda peaks2D:
                             map(lambda peaks1D:
                                 slice(*np.sort(peaks1D[0][peaks1D[1]['prominences'].argsort()[-2:][::-1]])),
                                 peaks2D),
                             peaks3D),
                         peaks)
            popts = map(lambda (I3D, V3D, s3D):
                        map(lambda (I2D, V2D, s2D):
                            map(lambda (I1D, V1D, s1D):
                                lin_fit(I1D[s1D], V1D[s1D]),
                                zip(I2D, V2D, s2D)),
                            zip(I3D, V3D, s3D)),
                        zip(I, V, slices))
            self.V_corr = np.array(map(lambda (I3D, V3D, popt3D):
                                       map(lambda (I2D, V2D, popt2D):
                                           map(lambda (I1D, V1D, popt1D):
                                               V1D - (popt1D[0] * I1D + popt1D[1]),
                                               zip(I2D, V2D, popt2D)),
                                           zip(I3D, V3D, popt3D)),
                                       zip(I, V, popts)))
            return self.V_corr
        else:
            raise ValueError('Scan dimension must be in {1, 2, 3}')

    def get_Ic_threshold(self, I=None, V=None, dVdI=None, threshold=20e-6, offset=None, Ir=False):
        """
        Get critical current values. These are considered as currents, where the voltage jumps beyond threshold ± <threshold> - <offset>.
        Parameters
        ----------
        I: numpy.array (optional)
            An N-dimensional array containing current values. Default is None that means self.I.
        V: numpy.array (optional)
            An N-dimensional array containing voltage values. Default is None that means self.V.
        threshold: float (optional)
            Threshold voltage that limits the superconducting branch. Default is 20e-6.
        offset: float (optional)
            Voltage offset that shifts the limits of the superconducting branch which is set by <threshold>. Default is None that means self.V_offset.
        Ir: bool (optional)
            Condition, if retrapping currents are returned, too. Default is False.
        Returns
        -------
        I_cs: numpy.array
            Critical current values.
        I_rs: numpy.array (optional)
            Retrapping current values.
        """
        if V is None:
            V = self.V
        if I is None:
            I = self.I
        if offset is None:
            if self.V_offset is None:
                offset = 0
            else:
                offset = self.V_offset
        if len(V.shape)-1 == 0: # single trace used for in situ fit
            with np.errstate(invalid='ignore'):  # raises warning due to np.nan
                mask = np.logical_and(V >= -threshold + offset, V <= threshold + offset)
            I_sc = np.copy(I)
            np.place(I_sc, np.logical_not(mask), np.nan)
            return np.nanmax(I_sc)
        ''' constant range via threshold (for JJ superconducting range via threshold voltage) '''
        with np.errstate(invalid='ignore'):  # raises warning due to np.nan
            mask = np.logical_and(V >= -threshold + offset, V <= threshold + offset)
        V_sc, I_sc = np.copy(V), np.copy(I)
        np.place(V_sc, np.logical_not(mask), np.nan)
        np.place(I_sc, np.logical_not(mask), np.nan)
        if self.sweeptype == 0:  # halfswing
            ''' critical current '''
            I_cs = np.array([np.nanmin(I_sc[0], axis=(self.scan_dim-1)), np.nanmax(I_sc[1], axis=(self.scan_dim-1))])
            ''' retrapping current '''
            I_rs = np.array([np.nanmax(I_sc[0], axis=(self.scan_dim-1)), np.nanmin(I_sc[1], axis=(self.scan_dim-1))])
        elif self.sweeptype == 1:  # 4 quadrants
            ''' critical current '''
            I_cs = np.array([np.nanmax(I_sc[0], axis=(self.scan_dim-1)), np.nanmin(I_sc[2], axis=(self.scan_dim-1))])
            ''' retrapping current '''
            I_rs = np.array([np.nanmax(I_sc[1], axis=(self.scan_dim-1)), np.nanmin(I_sc[3], axis=(self.scan_dim-1))])
        else:  # custom sweeptype
            ''' critical current '''
            I_cs = np.array([np.nanmax(I_sc, axis=(self.scan_dim-1)),
                             np.nanmin(I_sc, axis=(self.scan_dim-1))])
            ''' retrapping current '''
            if Ir:
                raise NotImplementedError('No algorithm implemented for custom sweeptype')
        if Ir:
            return [I_cs, I_rs]
        else:
            return I_cs

    def get_Ic_deriv(self, I=None, V=None, dVdI=None, Ir=False, tol_offset=20e-6, window=5, peak_finder=sig.find_peaks, **kwargs):
        """
        Gets critical current values using the numerical derivative dV/dI.
        Peaks in these data correspond to voltage jumps, are detected with a peak finding algorithm <peak_finder> and checked, whether the corresponding voltage jumps out or in the superconducting branch, that is identified as critical or retrapping current, respectively. Therefore the average of half the window below and above the peak is considered. The superconducting branch, in turn, is assumed as the voltage offset <self.V_offset> within the tolerance <tol_offset>.
        Parameters
        ----------
        I: numpy.array (optional)
            An N-dimensional array containing current values. Default is None that means self.I.
        V: numpy.array (optional)
            An N-dimensional array containing voltage values. Default is None that means self.V.
        dVdI: numpy.array (optional)
            An N-dimensional array containing differential resistance (dV/dI) values. Default is None that means self.dVdI.
        Ir: bool (optional)
            Condition, if retrapping currents are returned, too. Default is False
        tol_offset: float (optional)
            Voltage offset tolerance that limits the superconducting branch around the voltage offset <self.V_offset>. Default is 20e-6.
        window: int (optional)
            Window around the jump, where the voltage is evaluated and classified as 'superconducting branch'. Default is 5 that considers two values below and 2 values above the jump.
        peak_finder: function (optional)
            Peak finding algorithm. Default is scipy.signal.find_peaks.
        kwargs:
            Keyword arguments forwarded to the peak finding algorithm <peak_finder>.
        Returns
        -------
        I_cs: numpy.array
            Critical current values.
        I_rs: numpy.array (optional)
            Retrapping current values.
        properties: dict
            Properties of all found peaks (not only I_c and I_r, but also further jumps), such as corresponding currents, voltages, differential resistances, indices as well as returns of the used peak finding algorithm.
        Examples
        --------
        >>> I_cs, props = ivc.get_Ic_deriv(prominence=100)
        >>> if ivc.scan_dim == 1:
        >>>     Is = np.array(map(lambda p1D: p1D['I'], props))  # has shape (number of sweeps, number of peaks)
        >>> elif ivc.scan_dim == 2:
        >>>     Is = np.array(map(lambda p2D: map(lambda p1D: p1D['I'], p2D), props))  # has shape (number of sweeps, number of x-values, number of peaks)
        >>> elif ivc.scan_dim == 3:
        >>>     Is = np.array(map(lambda p3D: map(lambda p2D: map(lambda p1D: p1D['I'], p2D), p3D), props))  # has shape (number of sweeps, number of y-values, number of x-values, number of peaks)
        """
        def _peak_finder(x, **_kwargs):
            ans = peak_finder(x, **_kwargs)
            if np.array_equal(ans[0], []):  # no peaks found
                return [np.array([False]), {}]
            else:
                return ans
        if I is None:
            I = self.I
        if V is None:
            V = self.V
        if dVdI is None:
            dVdI = self.dVdI
        if peak_finder is sig.find_peaks and 'prominence' not in kwargs.keys():
            kwargs['prominence'] = 100
        if peak_finder is sig.find_peaks_cwt and 'widths' not in kwargs.keys():
            kwargs['widths'] = np.arange(10)
        if len(V.shape)-1 == 0: # single trace used for in situ fit
            peaks = _peak_finder(dVdI, **kwargs)
            try:
                return I[peaks[0]]
            except IndexError:  # if no peak found return np.nan
                return np.nan
        ''' peak detection in dV/dI '''
        if self.scan_dim == 1:
            peaks = np.array(map(lambda dVdI1D:
                                 _peak_finder(dVdI1D, **kwargs),
                                 dVdI))
        elif self.scan_dim == 2:
            peaks = np.array(map(lambda dVdI2D:
                                 map(lambda dVdI1D:
                                     _peak_finder(dVdI1D, **kwargs),
                                     dVdI2D),
                                 dVdI))
        elif self.scan_dim == 3:
            peaks = np.array(map(lambda dVdI3D:
                                 map(lambda dVdI2D:
                                     map(lambda dVdI1D:
                                         _peak_finder(dVdI1D, **kwargs),
                                         dVdI2D),
                                     dVdI3D),
                                 dVdI))
        else:
            raise ValueError('Scan dimension must be in {1, 2, 3}')
        return self._classify_jump(I=I, V=V, Y=dVdI, peaks=peaks, tol_offset=tol_offset, window=window, Ir=Ir)

    def get_Ic_dft(self, I=None, V=None, s=10, dVdI=None, Ir=False, tol_offset=20e-6, window=5, peak_finder=sig.find_peaks, **kwargs):
        """
        Gets critical current values using a discrete Fourier transform, a smoothed derivation in the frequency domain and an inverse Fourier transform.
        Therefore the voltage values are corrected by the linear offset slope, fast Fourier transformed to the frequency domain, multiplied with a Gaussian smoothed derivation function if*exp(-s*f^2) in the frequency domain and inversely fast Fourier transformed to the time domain. This corresponds to the convolution of the voltage values with the Gaussian smoothed derivation function in the time domain.
        Peaks in these data correspond to voltage jumps, are detected with a peak finding algorithm <peak_finder> and checked, whether the corresponding voltage jumps out or in the superconducting branch, that is identified as critical or retrapping current, respectively. Therefore the average of half the window below and above the peak is considered. The superconducting branch, in turn, is assumed as the voltage offset <self.V_offset> within the tolerance <tol_offset>.
        Parameters
        ----------
        I: numpy.array (optional)
            An N-dimensional array containing current values. Default is None that means self.I.
        V: numpy.array (optional)
            An N-dimensional array containing voltage values. Default is None that means self.V.
        s: float (optional)
            Smoothing factor of the derivative. Default is 10.
        Ir: bool (optional)
            Condition, if retrapping currents are returned, too. Default is False.
        tol_offset: float (optional)
            Voltage offset tolerance that limits the superconducting branch around the voltage offset <self.V_offset>. Default is 20e-6.
        window: int (optional)
            Window around the jump, where the voltage is evaluated and classified as 'superconducting branch'. Default is 5 that considers two values below and 2 values above the jump.
        peak_finder: function (optional)
            Peak finding algorithm. Default is scipy.signal.find_peaks
        kwargs:
            Keyword arguments forwarded to the peak finding algorithm <peak_finder>.
        Returns
        -------
        I_cs: numpy.array
            Critical current values.
        I_rs: numpy.array (optional)
            Retrapping current values.
        properties: dict
            Properties of all found peaks (not only I_c and I_r, but also further jumps), such as corresponding currents, voltages, differential resistances, indices as well as returns of the used peak finding algorithm.
        Examples
        --------
        >>> I_cs, props = ivc.get_Ic_dft(prominence=1e-5)
        >>> if ivc.scan_dim == 1:
        >>>     Is = np.array(map(lambda p1D: p1D['I'], props))  # has shape (number of sweeps, number of peaks)
        >>> elif ivc.scan_dim == 2:
        >>>     Is = np.array(map(lambda p2D: map(lambda p1D: p1D['I'], p2D), props))  # has shape (number of sweeps, number of x-values, number of peaks)
        >>> elif ivc.scan_dim == 3:
        >>>     Is = np.array(map(lambda p3D: map(lambda p2D: map(lambda p1D: p1D['I'], p2D), p3D), props))  # has shape (number of sweeps, number of y-values, number of x-values, number of peaks)
        """
        def _get_deriv_dft(V_corr):
            V_fft = np.fft.fft(V_corr)  # Fourier transform of V from time to frequency domain
            f = np.fft.fftfreq(V.shape[-1])  # frequency values
            kernel = 1j * np.fft.fft(f * np.exp(-s * f ** 2))  # smoothed derivation function, how it would look like in time domain (with which V is convolved in the time domain)
            V_fft_smooth = 1j * f * np.exp(-s * f ** 2) * V_fft  # Fourier transform of a Gaussian smoothed derivation of V in the frequency domain
            dV_smooth = np.fft.ifft(V_fft_smooth)  # inverse Fourier transform of the smoothed derivation of V from reciprocal to time domain
            return dV_smooth
        def _peak_finder(x, **_kwargs):
            ans = peak_finder(x, **_kwargs)
            if np.array_equal(ans[0], []):  # no peaks found
                return [np.array([False]), {}]
            else:
                return ans
        if I is None:
            I = self.I
        if V is None:
            V = self.V
        if peak_finder is sig.find_peaks and 'prominence' not in kwargs.keys():
            kwargs['prominence'] = 1e-5
        if peak_finder is sig.find_peaks_cwt and 'widths' not in kwargs.keys():
            kwargs['widths'] = np.arange(10)
        if len(V.shape)-1 == 0: # single trace used for in situ fit
            V_corr = V - np.linspace(start=V[0], stop=V[-1], num=V.shape[-1], axis=0)  # adjust offset slope
            dV_smooth = _get_deriv_dft(V_corr)
            peaks = _peak_finder(dV_smooth, **kwargs)
            print(len(peaks), peaks)
            try:
                return I[peaks[0]]
            except IndexError:  # if no peak found return np.nan
                return np.nan
        ''' differentiate and smooth in the frequency domain '''
        if self.scan_dim == 1:
            V_corr = V - np.linspace(start=V[:, 0], stop=V[:, -1], num=V.shape[-1], axis=1)  # adjust offset slope
        elif self.scan_dim == 2:
            V_corr = V - np.linspace(start=V[:, :, 0], stop=V[:, :, -1], num=V.shape[-1], axis=2)  # adjust offset slope
        elif self.scan_dim == 3:
            V_corr = V - np.linspace(start=V[:, :, :, 0], stop=V[:, :, :, -1], num=V.shape[-1], axis=3)  # adjust offset slope
        else:
            raise ValueError('Scan dimension must be in {1, 2, 3}')
        dV_smooth = _get_deriv_dft()
        ''' peak detection '''
        if self.scan_dim == 1:
            peaks = np.array(map(lambda dV_smooth1D:
                                 _peak_finder(np.abs(dV_smooth1D), **kwargs),
                                 dV_smooth))
        elif self.scan_dim == 2:
            peaks = np.array(map(lambda dV_smooth2D:
                                 map(lambda dV_smooth1D:
                                     _peak_finder(np.abs(dV_smooth1D), **kwargs),
                                     dV_smooth2D),
                                 dV_smooth))
        elif self.scan_dim == 3:
            peaks = np.array(map(lambda dV_smooth3D:
                                 map(lambda dV_smooth2D:
                                     map(lambda dV_smooth1D:
                                         _peak_finder(np.abs(dV_smooth1D), **kwargs),
                                         dV_smooth2D),
                                     dV_smooth3D),
                                 dV_smooth))
        else:
            raise ValueError('Scan dimension must be in {1, 2, 3}')
        return self._classify_jump(I=I, V=V, Y=dV_smooth, peaks=peaks, tol_offset=tol_offset, window=window, Ir=Ir)

    def _classify_jump(self, I, V, Y, peaks, tol_offset=20e-6, window=5, Ir=False):
        """
        Classifies voltage jumps as critical currents, retrapping currents of none of those.
        Parameters
        ----------
        I: numpy.array
            An N-dimensional array containing current values.
        V: numpy.array
            An N-dimensional array containing voltage values.
        Y: numpy.array
            An N-dimensional array containing data, whose peaks are already determined.
        peaks numpy.array
            An N-dimensional array containing indices and properties of peaks that are already determined, as obtained by e.g. scipy.signal.find_peaks()
        Ir: bool (optional)
            Condition, if retrapping currents are returned, too. Default is False
        tol_offset: float (optional)
            Voltage offset tolerance that limits the superconducting branch around the voltage offset <self.V_offset>. Default is 20e-6.
        window: int (optional)
            Window around the jump, where the voltage is evaluated and classified as 'superconducting branch'. Default is 5 that considers two values below and 2 values above the jump.
        Returns
        -------
        I_cs: numpy.array
            Critical current values.
        I_rs: numpy.array (optional)
            Retrapping current values.
        properties: dict
            Properties of all found peaks (not only I_c and I_r, but also further jumps), such as corresponding currents, voltages, differential resistances, indices as well as returns of the used peak finding algorithm.
        """
        Y_name = [key for key, val in sys._getframe().f_back.f_locals.items() if np.array_equal(val, Y)][0]
        if self.V_offset is None and self.sweeptype in [0, 1]:  # voltage offset to identify superconducting branch
            self.get_offset(x=V, y=I)
        if self.sweeptype == 0:  # halfswing
            V_c, I_c, peaks_c = np.copy(V), np.copy(I), np.copy(peaks)
            V_r, I_r, peaks_r = np.copy(V), np.copy(I), np.copy(peaks)
        elif self.sweeptype == 1:  # 4quadrants
            V_c, I_c, peaks_c = np.copy(V[::2]), np.copy(I[::2]), np.copy(peaks[::2])
            V_r, I_r, peaks_r = np.copy(V[1::2]), np.copy(I[1::2]), np.copy(peaks[1::2])
        else:  # custom sweeptype
            def f(ind1D, prop1D, I1D, V1D, Y1D):
                try:
                    return dict(prop1D.items() +
                                {k: v for k, v in zip(('I', 'V', Y_name, 'index'),
                                                      (I1D[ind1D], V1D[ind1D], Y1D[ind1D], ind1D))}.items())
                except IndexError:  # if no peak found return np.nan
                    return {k: v for k, v in zip(('I', 'V', Y_name, 'index'),
                                                 (np.array([np.nan]), np.array([np.nan]), np.array([np.nan]), np.array([np.nan])))}
            if self.scan_dim == 1:
                return np.array(map(lambda ((ind1D, prop1D), I1D, V1D, Y1D):
                                         f(ind1D, prop1D, I1D, V1D, Y1D),
                                         zip(peaks, I, V, Y)))
            elif self.scan_dim == 2:
                return np.array(map(lambda (peaks2D, I2D, V2D, Y2D):
                                    map(lambda ((ind1D, prop1D), I1D, V1D, Y1D):
                                        f(ind1D, prop1D, I1D, V1D, Y1D),
                                        zip(peaks2D, I2D, V2D, Y2D)),
                                    zip(peaks, I, V, Y)))
            elif self.scan_dim == 3:
                return np.array(map(lambda (peaks3D, I3D, V3D, Y3D):
                                    map(lambda (peaks2D, I2D, V2D, Y2D):
                                        map(lambda ((ind1D, prop1D), I1D, V1D, Y1D):
                                            f(ind1D, prop1D, I1D, V1D, Y1D),
                                            zip(peaks2D, I2D, V2D, Y2D)),
                                        zip(peaks3D, I3D, V3D, Y3D)),
                                    zip(peaks, I, V, Y)))
            else:
                raise NotImplementedError('No algorithm implemented for custom sweeptype and scan_dim > 3.')
        if self.scan_dim == 1:
            ''' critical current '''
            masks_c = np.array(map(lambda (V_c1D, peak1D):
                                   np.logical_and(np.logical_and((-tol_offset+self.V_offset) <= np.mean(V_c1D[peak1D[0]-np.tile([np.arange(int(window)/2)+1], [len(peak1D[0]), 1]).T], axis=0),
                                                                 np.mean(V_c1D[peak1D[0]-np.tile([np.arange(int(window)/2)+1], [len(peak1D[0]), 1]).T], axis=0) <= (+tol_offset+self.V_offset)),
                                                  np.logical_not(np.logical_and((-tol_offset+self.V_offset) <= np.mean(V_c1D[peak1D[0]+np.tile([np.arange(int(window)/2)+1], [len(peak1D[0]), 1]).T], axis=0),
                                                                                np.mean(V_c1D[peak1D[0]+np.tile([np.arange(int(window)/2)+1], [len(peak1D[0]), 1]).T], axis=0) <= (+tol_offset+self.V_offset)))),
                                   zip(V_c, peaks_c)))
            I_cs = np.array(map(lambda (I_c1D, peak1D, masks_c1D):
                                I_c1D[peak1D[0][masks_c1D][0]],
                                zip(I_c, peaks_c, masks_c)))
            ''' retrapping current '''
            masks_r = np.array(map(lambda (V_c1D, peak1D):
                                   np.logical_and(np.logical_not(np.logical_and((-tol_offset+self.V_offset) <= np.mean(V_c1D[peak1D[0]-np.tile([np.arange(int(window)/2)+1], [len(peak1D[0]), 1]).T], axis=0),
                                                                                np.mean(V_c1D[peak1D[0]-np.tile([np.arange(int(window)/2)+1], [len(peak1D[0]), 1]).T], axis=0) <= (+tol_offset+self.V_offset))),
                                                  np.logical_and((-tol_offset+self.V_offset) <= np.mean(V_c1D[peak1D[0]+np.tile([np.arange(int(window)/2)+1], [len(peak1D[0]), 1]).T], axis=0),
                                                                 np.mean(V_c1D[peak1D[0]+np.tile([np.arange(int(window)/2)+1], [len(peak1D[0]), 1]).T], axis=0) <= (+tol_offset+self.V_offset))),
                                   zip(V_r, peaks_r)))
            I_rs = np.array(map(lambda (I_r1D, peak1D, masks_r1D):
                                I_r1D[peak1D[0][masks_r1D][0]+1],
                                zip(I_r, peaks_r, masks_r)))
            ''' properties '''
            properties = map(lambda ((ind1D, prop1D), I1D, V1D, Y1D):
                             dict(prop1D.items() +
                                  {k: v for k, v in zip(('I', 'V', Y_name, 'index'),
                                                        (I1D[ind1D], V1D[ind1D], Y1D[ind1D], ind1D))}.items()),
                             zip(peaks, I, V, Y))
        elif self.scan_dim == 2:
            ''' critical current '''
            masks_c = np.array(map(lambda (V_c2D, peaks_c2D):
                                   map(lambda (V_c1D, peak1D): np.logical_and(np.logical_and((-tol_offset+self.V_offset) <= np.mean(V_c1D[peak1D[0]-np.tile([np.arange(int(window)/2)+1], [len(peak1D[0]), 1]).T], axis=0),
                                                                                             np.mean(V_c1D[peak1D[0]-np.tile([np.arange(int(window)/2)+1], [len(peak1D[0]), 1]).T], axis=0) <= (+tol_offset+self.V_offset)),
                                                                              np.logical_not(np.logical_and((-tol_offset+self.V_offset) <= np.mean(V_c1D[peak1D[0]+np.tile([np.arange(int(window)/2)+1], [len(peak1D[0]), 1]).T], axis=0),
                                                                                                            np.mean(V_c1D[peak1D[0]+np.tile([np.arange(int(window)/2)+1], [len(peak1D[0]), 1]).T], axis=0) <= (+tol_offset+self.V_offset)))),
                                       zip(V_c2D, peaks_c2D)),
                                   zip(V_c, peaks_c)))
            I_cs = np.array(map(lambda (I_c2D, peaks_c2D, masks_c2D):
                                map(lambda (I_c1D, peak1D, masks_c1D):
                                    I_c1D[peak1D[0][masks_c1D][0]],
                                    zip(I_c2D, peaks_c2D, masks_c2D)),
                                zip(I_c, peaks_c, masks_c)))
            ''' retrapping current '''
            masks_r = np.array(map(lambda (V_r2D, peaks_r2D):
                                   map(lambda (v, p): np.logical_and(np.logical_not(np.logical_and((-tol_offset+self.V_offset) <= np.mean(v[p[0]-np.tile([np.arange(int(window)/2)+1], [len(p[0]), 1]).T], axis=0),
                                                                                                   np.mean(v[p[0]-np.tile([np.arange(int(window)/2)+1], [len(p[0]), 1]).T], axis=0) <= (+tol_offset + self.V_offset))),
                                                                     np.logical_and((-tol_offset+self.V_offset) <= np.mean(v[p[0]+np.tile([np.arange(int(window)/2)+1], [len(p[0]), 1]).T], axis=0),
                                                                                    np.mean(v[p[0]+np.tile([np.arange(int(window)/2)+1], [len(p[0]), 1]).T], axis=0) <= (+tol_offset+self.V_offset))),
                                       zip(V_r2D, peaks_r2D)),
                                   zip(V_r, peaks_r)))
            I_rs = np.array(map(lambda (I_r2D, peaks_r2D, masks_r2D):
                                map(lambda (I_r1D, peaks_r1D, masks_r1D):
                                    I_r1D[peaks_r1D[0][masks_r1D][0]+1],
                                    zip(I_r2D, peaks_r2D, masks_r2D)),
                                zip(I_r, peaks_r, masks_r)))
            ''' properties '''
            properties = map(lambda (peaks2D, I2D, V2D, Y2D):
                             map(lambda ((ind1D, prop1D), I1D, V1D, Y1D):
                                 dict(prop1D.items() +
                                      {k: v for k, v in zip(('I', 'V', Y_name, 'index'),
                                                            (I1D[ind1D], V1D[ind1D], Y1D[ind1D], ind1D))}.items()),
                                 zip(peaks2D, I2D, V2D, Y2D)),
                             zip(peaks, I, V, Y))
        elif self.scan_dim == 3:
            ''' critical current '''
            masks_c = np.array(map(lambda (V_c3D, peaks_c3D):
                                   map(lambda (V_c2D, peaks_c2D):
                                       map(lambda (V_c1D, peak1D): np.logical_and(np.logical_and((-tol_offset+self.V_offset) <= np.mean(V_c1D[peak1D[0]-np.tile([np.arange(int(window)/2)+1], [len(peak1D[0]), 1]).T], axis=0),
                                                                                                 np.mean(V_c1D[peak1D[0]-np.tile([np.arange(int(window)/2)+1], [len(peak1D[0]), 1]).T], axis=0) <= (+tol_offset+self.V_offset)),
                                                                                  np.logical_not(np.logical_and((-tol_offset+self.V_offset) <= np.mean(V_c1D[peak1D[0]+np.tile([np.arange(int(window)/2)+1], [len(peak1D[0]), 1]).T], axis=0),
                                                                                                                np.mean(V_c1D[peak1D[0]+np.tile([np.arange(int(window)/2)+1], [len(peak1D[0]), 1]).T], axis=0) <= (+tol_offset+self.V_offset)))),
                                           zip(V_c2D, peaks_c2D)),
                                       zip(V_c3D, peaks_c3D)),
                                   zip(V_c, peaks_c)))
            I_cs = np.array(map(lambda (I_c3D, peaks_c3D, masks_c3D):
                                map(lambda (I_c2D, peaks_c2D, masks_c2D):
                                    map(lambda (I_c1D, peak1D, masks_c1D):
                                        I_c1D[peak1D[0][masks_c1D][0]],
                                        zip(I_c2D, peaks_c2D, masks_c2D)),
                                    zip(I_c3D, peaks_c3D, masks_c3D)),
                                zip(I_c, peaks_c, masks_c)))
            ''' retrapping current '''
            masks_r = np.array(map(lambda (V_r3D, peaks_r3D):
                                   map(lambda (V_r2D, peaks_r2D):
                                       map(lambda (v, p): np.logical_and(np.logical_not(np.logical_and((-tol_offset+self.V_offset) <= np.mean(v[p[0]-np.tile([np.arange(int(window)/2)+1], [len(p[0]), 1]).T], axis=0),
                                                                                                       np.mean(v[p[0]-np.tile([np.arange(int(window)/2)+1], [len(p[0]), 1]).T], axis=0) <= (+tol_offset + self.V_offset))),
                                                                         np.logical_and((-tol_offset+self.V_offset) <= np.mean(v[p[0]+np.tile([np.arange(int(window)/2)+1], [len(p[0]), 1]).T], axis=0),
                                                                                        np.mean(v[p[0]+np.tile([np.arange(int(window)/2)+1], [len(p[0]), 1]).T], axis=0) <= (+tol_offset+self.V_offset))),
                                           zip(V_r2D, peaks_r2D)),
                                       zip(V_r3D, peaks_r3D)),
                                   zip(V_r, peaks_r)))
            I_rs = np.array(map(lambda (I_r3D, peaks_r3D, masks_r3D):
                                map(lambda (I_r2D, peaks_r2D, masks_r2D):
                                    map(lambda (I_r1D, peak1D, masks_r1D):
                                        I_r1D[peak1D[0][masks_r1D][0]],
                                        zip(I_r2D, peaks_r2D, masks_r2D)),
                                    zip(I_r3D, peaks_r3D, masks_r3D)),
                                zip(I_r, peaks_r, masks_r)))
            ''' properties '''
            properties = map(lambda (peaks3D, I3D, V3D, Y3D):
                             map(lambda (peaks2D, I2D, V2D, Y2D):
                                 map(lambda ((ind1D, prop1D), I1D, V1D, Y1D):
                                     dict(prop1D.items() +
                                          {k: v for k, v in zip(('I', 'V', Y_name, 'index'),
                                                                (I1D[ind1D], V1D[ind1D], Y1D[ind1D], ind1D))}.items()),
                                     zip(peaks2D, I2D, V2D, Y2D)),
                                 zip(peaks3D, I3D, V3D, Y3D)),
                             zip(peaks, I, V, Y))
        else:
            raise ValueError('Scan dimension must be in {1, 2, 3}')
        if Ir:
            return I_cs, I_rs, properties
        else:
            return I_cs, properties


# two_port_switch_manual_client.py by JB@KIT 04/2015 jochen.braumueller@kit.edu
# client script to be run on the control machine (Raspberry Pi) to keep track of manual switching
# use: execute after starting the server

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import socket
import sys
from time import sleep

import RPi.GPIO as GPIO
GPIO.setmode(GPIO.BOARD)   #use RPi.GPIO layout

BUT1 = 13   #gpio27
BUT2 = 15   #gpio22   #button inputs
GPIO.setup(BUT1, GPIO.IN, pull_up_down=GPIO.PUD_DOWN)
GPIO.setup(BUT2, GPIO.IN, pull_up_down=GPIO.PUD_DOWN)   #connect buttons to +Vcc=3.3V

HOST, PORT = 'ip-address', 9988

def button1():
    return GPIO.input(BUT1) == 1
def button2():
    return GPIO.input(BUT2) == 1

if __name__ == "__main__":

    data = ''
    while True:
        if button1():
            data = 'switch 1'
        if button2():
            data = 'switch 2'
        if data != '':
            try:
                # Create a socket (SOCK_STREAM means a TCP socket)
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                # Connect to server and send data
                sock.connect((HOST, PORT))
                sock.sendall(data + "\n")
                # Receive data from the server and shut down
                received = sock.recv(1024)
            finally:
                print data
                data = ''
                sock.close()
                sleep(1)


from qkit.drivers.ZHInst_Abstract import ZHInst_Abstract
from zhinst.toolkit.control.drivers import PQSC as _PQSC


class ZHInst_PQSC(ZHInst_Abstract):
    def __init__(self, name, serialnumber, host="129.13.93.38",  **kwargs):
        super().__init__(name, **kwargs)
        self._pqsc = _PQSC(name, serialnumber, host=host)
        self._pqsc.setup()
        self._pqsc.connect_device()
        
        # Iterate node tree for readable entries and hook them into QKit
        self.blacklist = ["system_fwlog", "features_code"]
        self.mount_api("node_dump_shfsg.txt", self._pqsc.nodetree)

        # Register readout methods



# load_awg.py
# adapted from the old load_awg, started by M. Jerger and adapted by AS, JB
# and now adjusted for the virutal_awg by TW
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# import qkit  # ToDO: flow comments?
import numpy as np
import logging
from qkit.gui.notebook.Progress_Bar import Progress_Bar
import gc


def _adjust_wfs_for_tabor(wf1, wf2, ro_index, chpair, segment, sample):
    """
    This function simply adjust the waveforms, coming from the virtual awg, to fit the requirements of the
    Tabor awg
    """
    divisor = 16
    readout_ind = int(ro_index[segment] + int(sample.clock * sample.readout_delay))
    # Adjust length to fit marker at arbitrary positions
    if readout_ind + 10 > len(wf1):
        marker1 = np.zeros(readout_ind + 10)
        marker1[readout_ind:readout_ind + 10] = 1
        wf1 = np.append(wf1, np.zeros(readout_ind + 10 - len(wf1)))
        wf2 = np.append(wf2, np.zeros(readout_ind + 10 - len(wf2)))
    elif readout_ind < 0:
        marker1 = np.append(np.zeros(np.abs(readout_ind)), np.zeros_like(wf1))
        marker1[0:10] = 1
        wf1 = np.append(np.zeros(np.abs(readout_ind)), wf1)
        wf2 = np.append(np.zeros(np.abs(readout_ind)), wf2)
    else:
        marker1 = np.zeros_like(wf1)
        marker1[readout_ind:readout_ind + 10] = 1
    # for single channel pulses
    if len(wf1) > len(wf2):
        wf2 = np.append(wf2, np.zeros(len(wf1)-len(wf2)))
    
    # all wfms start and end with 0
    wf1 = np.append(0, wf1)
    wf1 = np.append(wf1, 0)
    wf2 = np.append(0, wf2)
    wf2 = np.append(wf2, 0)
    marker1 = np.append(0, marker1)
    marker1 = np.append(marker1, 0)
        
    end_zeros = len(wf1) % divisor
    # minimum waveform is 192 points, handled by the Tabor driver
    if end_zeros != 0:
        wf1 = np.append(np.zeros(divisor - end_zeros), wf1)
        wf2 = np.append(np.zeros(divisor - end_zeros), wf2)
        marker1 = np.append(np.zeros(divisor - end_zeros), marker1)
    sample.awg.wfm_send2(wf1, wf2, marker1, marker1, chpair * 2 - 1, segment + 1)


def load_tabor(channel_sequences, ro_index, sample, reset=True, show_progress_bar=True):
    """
    This function takes the data, coming from virtual awg, and loads them into the awg
    :param channel_sequences: This must be a list of list, i.e., a list of channels each containing the sequences
    :param ro_index: coming from virtual awg, for each sequence there is a ro_index letting the awg know, when to
                     send the trigger.
    :param sample: you should know this
    :param reset: simply sets the awg_channel active, probably not needed
    :param show_progress_bar: enables the progress bar
    :return:
    """
    awg = sample.awg
    awg.clear_waveforms()
    number_of_channels = 0
    complex_channel = []
    for chan in channel_sequences:
        if True in (s.dtype == np.complex128 for s in chan):
            number_of_channels += 2
            complex_channel.append(True)
        else:
            number_of_channels += 1
            complex_channel.append(False)
    if number_of_channels > awg._numchannels:
        raise ValueError('more sequences than channels')
    # reordering channels if necessary
    if number_of_channels > 2:
        if complex_channel[:2] == [False, True]:
            logging.warning('Channels reordered! Please do not split complex channels on two different channelpairs.'
                            'Complex channel is on chpair 1')
            channel_sequences[:2] = [channel_sequences[1], channel_sequences[0]]
            complex_channel[:2] = [True, False]
    #qkit.flow.start()

    if reset:
        _reset(awg, number_of_channels, ro_index)
    # Loading the waveforms into the AWG, differentiating between all cases
    if show_progress_bar:
        p = Progress_Bar(len(ro_index), 'Load AWG')

    if number_of_channels == 1:
        for j, seq in enumerate(channel_sequences[0]):
            _adjust_wfs_for_tabor(seq, [0], ro_index, 1, j, sample)
            if show_progress_bar:
                p.iterate()

    elif number_of_channels == 2:
        if complex_channel[0]:
            for j, seq in enumerate(channel_sequences[0]):
                _adjust_wfs_for_tabor(seq.real, seq.imag, ro_index, 1, j, sample)
                if show_progress_bar:
                    p.iterate()
        else:
            for j, seq in enumerate(zip(channel_sequences[0], channel_sequences[1])):
                _adjust_wfs_for_tabor(seq[0], seq[1], ro_index, 1, j, sample)
                if show_progress_bar:
                    p.iterate()

    elif number_of_channels == 3:
        if complex_channel[0]:
            for j, seq in enumerate(zip(channel_sequences[0], channel_sequences[1])):
                _adjust_wfs_for_tabor(seq[0].real, seq[0].imag, ro_index, 1, j, sample)
                _adjust_wfs_for_tabor(seq[1], [0], ro_index, 2, j, sample)
                if show_progress_bar:
                    p.iterate()
        else:
            for j, seq in enumerate(zip(channel_sequences[0], channel_sequences[1], channel_sequences[2])):
                _adjust_wfs_for_tabor(seq[0], seq[1], ro_index, 1, j, sample)
                _adjust_wfs_for_tabor(seq[2], [0], ro_index, 2, j, sample)
                if show_progress_bar:
                    p.iterate()

    else:  # 4 channels
        if complex_channel == [True, True]:
            for j, seq in enumerate(zip(channel_sequences[0], channel_sequences[1])):
                _adjust_wfs_for_tabor(seq[0].real, seq[0].imag, ro_index, 1, j, sample)
                _adjust_wfs_for_tabor(seq[1].real, seq[1].imag, ro_index, 2, j, sample)
                if show_progress_bar:
                    p.iterate()
        elif complex_channel == [True, False, False]:
            for j, seq in enumerate(zip(channel_sequences[0], channel_sequences[1], channel_sequences[2])):
                _adjust_wfs_for_tabor(seq[0].real, seq[0].imag, ro_index, 1, j, sample)
                _adjust_wfs_for_tabor(seq[1], seq[2], ro_index, 2, j, sample)
                if show_progress_bar:
                    p.iterate()
        elif complex_channel == [False, False, True]:
            for j, seq in enumerate(zip(channel_sequences[0], channel_sequences[1], channel_sequences[2])):
                _adjust_wfs_for_tabor(seq[0], seq[1], ro_index, 1, j, sample)
                _adjust_wfs_for_tabor(seq[2].real, seq[2].imag, ro_index, 2, j, sample)
                if show_progress_bar:
                    p.iterate()
        else:
            for j, seq in enumerate(zip(channel_sequences[0], channel_sequences[1],
                                        channel_sequences[2], channel_sequences[3])):
                _adjust_wfs_for_tabor(seq[0], seq[1], ro_index, 1, j, sample)
                _adjust_wfs_for_tabor(seq[2], seq[3], ro_index, 2, j, sample)
                if show_progress_bar:
                    p.iterate()

    gc.collect()

    if number_of_channels <= 2:
        num_channels = [1, 2]
    else:
        num_channels = [1, 2, 3, 4]
    return np.all([awg.get('ch%i_status' % i) for i in num_channels])


def _reset(awg, num_channels, ro_index):
    awg.set('p%i_runmode' % 1, 'SEQ')
    awg.define_sequence(1, len(ro_index))
    awg.set_ch1_status(True)
    awg.set_ch2_status(True)
    awg.set_ch1_offset(0)
    awg.set_ch2_offset(0)
    awg.set_ch1_amplitude(2)
    awg.set_ch2_amplitude(2)
    awg.set_ch1_marker_output(True)
    awg.set_ch1_marker_high(1)
    awg.set_ch2_marker_output(True)
    awg.set_ch2_marker_high(1)
    if num_channels > 2:
        awg.set('p%i_runmode' % 2, 'SEQ')
        awg.define_sequence(3, len(ro_index))
        awg.set_ch3_status(True)
        awg.set_ch4_status(True)
        awg.set_ch3_offset(0)
        awg.set_ch4_offset(0)
        awg.set_ch3_amplitude(2)
        awg.set_ch4_amplitude(2)
        awg.set_ch3_marker_output(True)
        awg.set_ch3_marker_high(1)
        awg.set_ch4_marker_output(True)
        awg.set_ch4_marker_high(1)



# fridge_monitoring.py by JB@KIT 06/2014
# script for monitoring UFO temperature
# edits 08/2015, 05/2017
# the critical temperature in mK is passed as an argument

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from sendMail import *
from tip_client import tip_client
import time
from numpy import mean
from numpy import abs
import sys


def _sendTemp(temp):
    
    strTime = time.strftime('%x %X')

    s = SendMail("UFO@kit.edu",strTime+" UFO monitoring")   #sender, subject
    
    s.setRecipients("")
    #s.add_Line("Hello, I am feeling great. Cool gases. Nice pressures.")
    s.add_Line(strTime)
    s.add_Line("Current base temperature: " + str(temp) + " mK")
    s.add_Line("")
    s.add_Line("Please reflect the above given value critically and make sure to interpret it in the right way.")
    s.add_Line("")
    s.add_Line("Best, UFO")
    
    s.sendEMail()
    print "Mail sent to recipients. Current base temperature: " + str(temp) + "."
    
    
def _getTemp(t):
    try:
        return round(float(t.r_get_T())*1000,2)   #T in mK
    except Exception as m:
        print 'Could not retrieve temperature', m
        return 0
    
def _parameters_critical(t,b,temp_critical):
    #returns True if a message is to be sent, False otherwise
    if b_mean > temp_critical and abs(b[-1]-b[-2]) < 15:   #if mean value of last [10] minutes > 35mK and the increase smooth
        if b[-2] > b[-1] or b[-1] < temp_critical:   #if temperature is decreasing
            return False
        else:
            return True
    else:
        return False


if __name__ == "__main__":
    
    t = tip_client('tip',address='pi-us74')   #tip raspberry
    
    # readout critical temperature argument
    if len(sys.argv) == 1:   #no argument passed -> set default 35mK
        temp_critical = 35
    else:
        temp_critical = float(sys.argv[1])   #in mK
    
    buffer = []
    for i in range(10):
        buffer.append(18.0)
    m_sent = False
    while True:
    time.sleep(59)
    if int(time.strftime('%M')) == 0:   #every full hour
        m_sent = False
    
    buffer.append(_getTemp(t))
    buffer = buffer[1:]
    b_mean = mean(buffer)
    print b_mean
    if _parameters_critical(t,buffer,temp_critical) and m_sent == False:   #if base T > 25mK
        _sendTemp(_getTemp(t))
            m_sent = True
            